(************************************************************************************
Filename    :   Win32_DirectXAppUtil.h
Content     :   D3D11 application/Window setup functionality for RoomTiny
Created     :   October 20th, 2014
Author      :   Tom Heath
Copyright   :   Copyright 2014 Oculus, Inc. All Rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
unit UDirectXAppUtil;

interface

uses
  Winapi.Windows, Winapi.Messages, Winapi.D3D11, Winapi.D3D10, Winapi.DXGI,
  Winapi.DxgiFormat, Winapi.DxgiType, Winapi.D3DCommon, System.SysUtils,
  System.Types, System.Math, System.Math.Vectors, System.Generics.Collections,
  OVR_CAPI, UCommon;

type
  TDepthBuffer = class
  public
    TexDsv: ID3D11DepthStencilView;
    constructor Create(const Device: ID3D11Device; sizeW, sizeH: Integer; sampleCount: Integer = 1);
  end;

  TDataBuffer = class
  private
    D3DBuffer: ID3D11Buffer;
    Size: NativeInt;
  public
    constructor Create(const Device: ID3D11Device; use: D3D11_BIND_FLAG;
      const buffer: Pointer; size: NativeInt);

    property Buffer: ID3D11Buffer read D3DBuffer;
  end;

  TMainLoopFunc = function(ARetryCreate: Boolean): Boolean;

  TDirectX11 = class
  private
  const
    // Fixed size buffer for shader constants, before copied into buffer
    UNIFORM_DATA_SIZE = 2000;
  var
    class function WindowProc(hWnd: THandle; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall; static;
  public
    Window: THandle;
    Running: Boolean;
    Key: array[#0..#255] of Boolean;
    KeyDown: array[#0..#255] of Boolean;
    WinSizeW: Integer;
    WinSizeH: Integer;
    Device: ID3D11Device;
    Context: ID3D11DeviceContext;
    SwapChain: IDXGISwapChain;
    MainDepthBuffer: TDepthBuffer;
    BackBuffer: ID3D11Texture2D;
    BackBufferRT: ID3D11RenderTargetView;
    UniformData: array[0..UNIFORM_DATA_SIZE-1] of Byte;
    UniformBufferGen: TDataBuffer;
    destructor Destroy; override;
    function InitWindow(const title: string): Boolean;
    procedure CloseWindow;
    function InitDevice(vpW, vpH: Integer; const pLuid: ovrGraphicsLuid; windowed: Boolean = True): Boolean;
    procedure SetAndClearRenderTarget(rendertarget: ID3D11RenderTargetView;
      depthbuffer: TDepthBuffer; R: Single = 0; G: Single = 0; B: Single = 0;
      A: Single = 0);
    procedure SetViewport(vpX, vpY, vpW, vpH: Single);
    function HandleMessages: Boolean;
    procedure Run(const MainLoop: TMainLoopFunc);
    procedure ReleaseDevice;
  end;

  TDWORDArray = array[0..0] of DWORD;
  PDWORDArray = ^TDWORDArray;

  TTexture = class
  public
  const
    AUTO_WHITE = 1;
    AUTO_WALL = 2;
    AUTO_FLOOR = 3;
    AUTO_CEILING = 4;
    AUTO_GRID = 5;
    AUTO_GRADE_256 = 6;
  var
    Tex: ID3D11Texture2D;
    TexSv: ID3D11ShaderResourceView;
    TexRtv: ID3D11RenderTargetView;
    SizeW, SizeH, MipLevels: Integer;
    constructor Create(sizeW, sizeH: Integer; rendertarget: Boolean;
      mipLevels: Integer = 1; sampleCount: Integer = 1); overload;
    constructor Create(rendertarget: Boolean; sizeW, sizeH: Integer;
      autoFillData: Integer = 0; sampleCount: Integer = 1); overload;
    procedure Init(sizeW, sizeH: Integer; rendertarget: Boolean; mipLevels, sampleCount: Integer);
    procedure FillTexture(pix: PDWORD);
    class procedure ConvertToSRGB(linear: PDWORD); static;
    procedure AutoFillTexture(autoFillData: Integer);
  end;

  TMaterial = class
  public
  const
    MAT_WRAP = 1;
    MAT_WIRE = 2;
    MAT_ZALWAYS = 4;
    MAT_NOCULL = 8;
    MAT_TRANS = 16;
    DEFAULT_FLAGS = MAT_WRAP or MAT_TRANS;
  private
  var
    VertexShader, VertexShaderInstanced: ID3D11VertexShader;
    PixelShader: ID3D11PixelShader;
    Tex: TTexture;
    InputLayout: ID3D11InputLayout;
    VertexSize: UINT;
    SamplerState: ID3D11SamplerState;
    Rasterizer: ID3D11RasterizerState;
    DepthState: ID3D11DepthStencilState;
    BlendState: ID3D11BlendState;
  public
    constructor Create(t: TTexture; flags: DWORD = DEFAULT_FLAGS;
      vertexDesc: PD3D11_INPUT_ELEMENT_DESC = nil; numVertexDesc: Integer = 3;
      vertexShader: PAnsiChar = nil; pixelShader: PAnsiChar = nil; vSize: Integer = 24);
    destructor Destroy; override;
  end;

  PVertex = ^TVertex;
  TVertex = packed record
  public
    Pos: TPoint3D;
    C: DWORD;
    U, V: Single;
    class function Create(pos: TPoint3D; c: DWORD; u, v: Single): TVertex; static;
  end;

  TTriangleSet = record
  private
    FVertices: TArray<TVertex>;
    FIndices: TArray<SmallInt>;
    function GetNumVertices: Integer;
    function GetNumIndices: Integer;
  public
    procedure AddQuad(const v0, v1, v2, v3: TVertex);
    procedure AddTriangle(const v0, v1, v2: TVertex);
    function ModifyColor(c: DWORD; pos: TPoint3D): DWORD;
    procedure AddSolidColorBox(x1, y1, z1, x2, y2, z2: Single; c: DWORD);

    property Vertices: TArray<TVertex> read FVertices write FVertices;
    property Indices: TArray<SmallInt> read FIndices write FIndices;
    property NumVertices: Integer read GetNumVertices;
    property NumIndices: Integer read GetNumIndices;
  end;

  TRenderItem = class
  protected
    FPos: TPoint3D;
    FRot: TQuaternion3D;
    procedure Update; virtual;
  public
    procedure Render(const projView: TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean); virtual; abstract;
    procedure RenderInstanced(const viewProjMats: array of TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean); virtual; abstract;

    property Pos: TPoint3D read FPos write FPos;
    property Rot: TQuaternion3D read FRot write FRot;
  end;

  TModel = class(TRenderItem)
  protected
    FFill: TMaterial;
    FVertexBuffer: TDataBuffer;
    FIndexBuffer: TDataBuffer;
    FNumIndices: Integer;
    FVisible: Boolean;
  public
    constructor Create(const t: TTriangleSet; argPos: TPoint3D; argRot: TQuaternion3D; argFill: TMaterial); overload;
    constructor Create(mat: TMaterial; minx, miny, maxx, maxy: Single; zDepth: Single = 0); overload;
    destructor Destroy; override;
    procedure Init(const t: TTriangleSet);
    procedure Render(const projView: TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean); override;
    procedure RenderInstanced(const viewProjMats: array of TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean); override;

    property Fill: TMaterial read FFill write FFill;
    property Visible: Boolean read FVisible write FVisible;
  end;

  TScene = class
  private
    FModels: TObjectList<TRenderItem>;
  public
    constructor Create(includeIntensiveGPUobject: Boolean);
    destructor Destroy; override;
    procedure Add(n: TRenderItem);
    procedure Render(const projView: TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean);
    procedure RenderInstanced(const projViews: array of TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean);
    procedure Init(includeIntensiveGPUobject: Boolean);
    procedure Update;

    property Models: TObjectList<TRenderItem> read FModels;
  end;

  TCamera = class
  private
    FPos: TPoint3D;
    FRot: TQuaternion3D;
  public
    constructor Create(const APos: TPoint3D; const ARot: TQuaternion3D);
    function GetViewMatrix: TMatrix3D;

    property Pos: TPoint3D read FPos write FPos;
    property Rot: TQuaternion3D read FRot write FRot;
  end;

var
  // global DX11 state
  DIRECTX: TDirectX11;

procedure VALIDATE(x: Boolean; const msg: string);

implementation

function D3DCompile(pSrcData: LPCVOID; SrcDataSize: NativeUInt;
  pSourceName: LPCSTR; pDefines: PD3D10_ShaderMacro;
  pInclude: ID3D10Include; pEntrypoint, pTarget: LPCSTR; Flags1, Flags2: UINT;
  out ppCode, ppErrorMsgs: ID3D10Blob): HRESULT; stdcall; external 'd3dcompiler_47.dll';

procedure VALIDATE(x: Boolean; const msg: string);
begin
  if not x then
  begin
    MessageBox(0, PChar(msg), 'OculusRoomTiny', MB_ICONERROR or MB_OK);
    ExitCode := -1;
    Halt;
  end;
end;

{ TDepthBuffer }

constructor TDepthBuffer.Create(const Device: ID3D11Device; sizeW, sizeH: Integer; sampleCount: Integer = 1);
var
  format: DXGI_FORMAT;
  dsDesc: D3D11_TEXTURE2D_DESC;
  Tex: ID3D11Texture2D;
begin
  inherited Create;

  format := DXGI_FORMAT_D32_FLOAT;
  dsDesc.Width := sizeW;
  dsDesc.Height := sizeH;
  dsDesc.MipLevels := 1;
  dsDesc.ArraySize := 1;
  dsDesc.Format := format;
  dsDesc.SampleDesc.Count := sampleCount;
  dsDesc.SampleDesc.Quality := 0;
  dsDesc.Usage := D3D11_USAGE_DEFAULT;
  dsDesc.CPUAccessFlags := 0;
  dsDesc.MiscFlags := 0;
  dsDesc.BindFlags := D3D11_BIND_DEPTH_STENCIL;
  Device.CreateTexture2D(dsDesc, nil, Tex);
  Device.CreateDepthStencilView(Tex, nil, TexDsv);
end;

{ TDataBuffer }

constructor TDataBuffer.Create(const Device: ID3D11Device; use: D3D11_BIND_FLAG;
  const buffer: Pointer; size: NativeInt);
var
  desc: D3D11_BUFFER_DESC;
  sr: D3D11_SUBRESOURCE_DATA;
begin
  inherited Create;

  Self.Size := size;
  FillChar(desc, SizeOf(desc), 0);
  desc.Usage := D3D11_USAGE_DYNAMIC;
  desc.CPUAccessFlags := D3D11_CPU_ACCESS_WRITE;
  desc.BindFlags := use;
  desc.ByteWidth := UINT(size);
  sr.pSysMem := buffer;
  sr.SysMemSlicePitch := 0;
  sr.SysMemPitch := 0;
  if buffer <> nil then
    Device.CreateBuffer(desc, @sr, D3DBuffer)
  else
    Device.CreateBuffer(desc, nil, D3DBuffer);
end;

{ TDirectX11 }

class function TDirectX11.WindowProc(hWnd: THandle; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  p: TDirectX11;
begin
  p := TDirectX11(GetWindowLongPtr(hWnd, 0));
  case Msg of
    WM_KEYDOWN:
    begin
      p.Key[Chr(wParam)] := True;
      if lParam and $40000000 = 0 then
        p.KeyDown[Chr(wParam)] := True;
    end;
    WM_KEYUP:
      p.Key[Chr(wParam)] := False;
    WM_SYSCOMMAND:
    begin
      if wParam = SC_KEYMENU then
        Exit(0);
      Exit(DefWindowProcW(hWnd, Msg, wParam, lParam));
    end;
    WM_DESTROY:
      p.Running := False;
  else
    Exit(DefWindowProcW(hWnd, Msg, wParam, lParam));
  end;
  if (p.Key['Q'] and p.Key[Chr(VK_CONTROL)]) or p.Key[Chr(VK_ESCAPE)] then
  begin
    p.Running := False;
  end;
  Result := 0;
end;

destructor TDirectX11.Destroy;
begin
  ReleaseDevice;
  CloseWindow;

  inherited;
end;

function TDirectX11.InitWindow(const title: string): Boolean;
var
  wc: WNDCLASSW;
begin
  Running := True;

  FillChar(wc, SizeOf(wc), 0);
  wc.lpszClassName := 'App';
  wc.style := CS_OWNDC;
  wc.lpfnWndProc := @WindowProc;
  wc.cbWndExtra := SizeOf(Self);
  RegisterClassW(wc);

  // adjust the window size and show at InitDevice time
  Window := CreateWindow(wc.lpszClassName, PChar(title), WS_OVERLAPPEDWINDOW, 0, 0, 0, 0, 0, 0, HInstance, nil);
  if Window = 0 then Exit(False);

  SetWindowLongPtr(Window, 0, LONG_PTR(Self));

  Result := True;
end;

procedure TDirectX11.CloseWindow;
begin
  if Window <> 0 then
  begin
    DestroyWindow(Window);
    Window := 0;
    UnregisterClass('App', HInstance);
  end;
end;

function TDirectX11.InitDevice(vpW, vpH: Integer; const pLuid: TOvrGraphicsLuid; windowed: Boolean = True): Boolean;
const
  flags: UINT = SWP_NOMOVE or SWP_NOZORDER or SWP_SHOWWINDOW;
var
  size: TRect;
  DXGIFactory: IDXGIFactory;
  hr: HRESULT;
  Adapter: IDXGIAdapter;
  iAdapter: UINT;
  adapterDesc: DXGI_ADAPTER_DESC;
  DriverType: D3D_DRIVER_TYPE;
  scDesc: DXGI_SWAP_CHAIN_DESC;
  DXGIDevice1: IDXGIDevice1;
begin
  WinSizeW := vpW;
  WinSizeH := vpH;

  size := Rect(0, 0, vpW, vpH);
  AdjustWindowRect(size, WS_OVERLAPPEDWINDOW, False);
  if not SetWindowPos(Window, 0, 0, 0, size.Width, size.Height, flags) then
    Exit(False);

  DXGIFactory := nil;
  hr := CreateDXGIFactory1(IDXGIFactory, DXGIFactory);
  VALIDATE(hr = ERROR_SUCCESS, 'CreateDXGIFactory1 failed');

  Adapter := nil;
  iAdapter := 0;
  while DXGIFactory.EnumAdapters(iAdapter, Adapter) <> DXGI_ERROR_NOT_FOUND do
  begin
    Adapter.GetDesc(adapterDesc);
    if CompareMem(@adapterDesc.AdapterLuid, @pLuid, SizeOf(LUID)) then
      Break;
    Inc(iAdapter);
  end;

  if Adapter <> nil then
    DriverType := D3D_DRIVER_TYPE_UNKNOWN
  else
    DriverType := D3D_DRIVER_TYPE_HARDWARE;
  hr := D3D11CreateDevice(Adapter, DriverType, 0, 0, nil, 0, D3D11_SDK_VERSION, Device, D3D_FEATURE_LEVEL(nil^), Context);
  Adapter := nil;
  VALIDATE(hr = ERROR_SUCCESS, 'D3D11CreateDevice failed');

  // Create swap chain
  FillChar(scDesc, SizeOf(scDesc), 0);
  scDesc.BufferCount := 2;
  scDesc.BufferDesc.Width := WinSizeW;
  scDesc.BufferDesc.Height := WinSizeH;
  scDesc.BufferDesc.Format := DXGI_FORMAT_R8G8B8A8_UNORM;
  scDesc.BufferDesc.RefreshRate.Denominator := 1;
  scDesc.BufferUsage := DXGI_USAGE_RENDER_TARGET_OUTPUT;
  scDesc.OutputWindow := Window;
  scDesc.SampleDesc.Count := 1;
  scDesc.Windowed := windowed;
  scDesc.SwapEffect := DXGI_SWAP_EFFECT_SEQUENTIAL;
  hr := DXGIFactory.CreateSwapChain(Device, scDesc, SwapChain);
  DXGIFactory := nil;
  VALIDATE(hr = ERROR_SUCCESS, 'CreateSwapChain failed');

  // Create backbuffer
  SwapChain.GetBuffer(0, ID3D11Texture2D, BackBuffer);
  hr := Device.CreateRenderTargetView(BackBuffer, nil, BackBufferRT);
  VALIDATE(hr = ERROR_SUCCESS, 'CreateRenderTargetView failed');

  // Main depth buffer
  MainDepthBuffer := TDepthBuffer.Create(Device, WinSizeW, WinSizeH);
  Context.OMSetRenderTargets(1, BackBufferRT, MainDepthBuffer.TexDsv);

  // Buffer for shader constants
  UniformBufferGen := TDataBuffer.Create(Device, D3D11_BIND_CONSTANT_BUFFER, nil, UNIFORM_DATA_SIZE);
  Context.VSSetConstantBuffers(0, 1, UniformBufferGen.D3DBuffer);

  // Set max frame latency to 1
  DXGIDevice1 := nil;
  hr := Device.QueryInterface(IDXGIDevice1, DXGIDevice1);
  VALIDATE(hr = ERROR_SUCCESS, 'QueryInterface failed');
  DXGIDevice1.SetMaximumFrameLatency(1);
  DXGIDevice1 := nil;

  Result := True;
end;

procedure TDirectX11.SetAndClearRenderTarget(rendertarget: ID3D11RenderTargetView;
  depthbuffer: TDepthBuffer; R: Single = 0; G: Single = 0; B: Single = 0;
  A: Single = 0);
var
  black: TFourSingleArray;
begin
  // Important that alpha=0, if want pixels to be transparent, for manual layers
  black[0] := R;
  black[1] := G;
  black[2] := B;
  black[3] := A;
  if depthbuffer <> nil then
    Context.OMSetRenderTargets(1, rendertarget, depthbuffer.TexDsv)
  else
    Context.OMSetRenderTargets(1, rendertarget, nil);
  Context.ClearRenderTargetView(rendertarget, black);
  if depthbuffer <> nil then
    Context.ClearDepthStencilView(depthbuffer.TexDsv, D3D11_CLEAR_DEPTH or D3D11_CLEAR_STENCIL, 1, 0);
end;

procedure TDirectX11.SetViewport(vpX, vpY, vpW, vpH: Single);
var
  D3Dvp: D3D11_VIEWPORT;
begin
  D3Dvp.Width := vpW;    D3Dvp.Height := vpH;
  D3Dvp.MinDepth := 0;   D3Dvp.MaxDepth := 1;
  D3Dvp.TopLeftX := vpX; D3Dvp.TopLeftY := vpY;
  Context.RSSetViewports(1, @D3Dvp);
end;

function TDirectX11.HandleMessages: Boolean;
var
  msg: tagMSG;
begin
  FillChar(KeyDown, SizeOf(KeyDown), 0);
  while PeekMessage(msg, 0, 0, 0, PM_REMOVE) do
  begin
    TranslateMessage(msg);
    DispatchMessage(msg);
  end;
  // This is to provide a means to terminate after a maximum number of frames
  // to facilitate automated testing
{$IFDEF MAX_FRAMES_ACTIVE}
  if maxFrames > 0 then
  begin
    Dec(maxFrames);
    if maxFrames <= 0 then
      Running := False;
  end;
{$ENDIF}
  Result := Running;
end;

procedure TDirectX11.Run(const MainLoop: TMainLoopFunc);
begin
  // false => just fail on any error
  VALIDATE(MainLoop(False), 'Oculus Rift not detected.');
  while HandleMessages do
  begin
    // true => we'll attempt to retry for ovrError_DisplayLost
    if not MainLoop(True) then
      Break;
    // Sleep a bit before retrying to reduce CPU load while the HMD is disconnected
    Sleep(10);
  end;
end;

procedure TDirectX11.ReleaseDevice;
begin
  BackBuffer := nil;
  BackBufferRT := nil;
  if SwapChain <> nil then
  begin
    SwapChain.SetFullscreenState(False, nil);
    SwapChain := nil;
  end;
  Context := nil;
  Device := nil;
  FreeAndNil(MainDepthBuffer);
  FreeAndNil(UniformBufferGen);
end;

{ TTexture }

procedure TTexture.Init(sizeW, sizeH: Integer; rendertarget: Boolean;
  mipLevels, sampleCount: Integer);
var
  dsDesc: D3D11_TEXTURE2D_DESC;
begin
  Self.SizeW := sizeW;
  Self.SizeH := sizeH;
  Self.MipLevels := mipLevels;

  dsDesc.Width := SizeW;
  dsDesc.Height := SizeH;
  dsDesc.MipLevels := MipLevels;
  dsDesc.ArraySize := 1;
  dsDesc.Format := DXGI_FORMAT_R8G8B8A8_UNORM;
  dsDesc.SampleDesc.Count := sampleCount;
  dsDesc.SampleDesc.Quality := 0;
  dsDesc.Usage := D3D11_USAGE_DEFAULT;
  dsDesc.CPUAccessFlags := 0;
  dsDesc.MiscFlags := 0;
  dsDesc.BindFlags := D3D11_BIND_SHADER_RESOURCE;
  if rendertarget then
    dsDesc.BindFlags := dsDesc.BindFlags or D3D11_BIND_RENDER_TARGET;

  DIRECTX.Device.CreateTexture2D(dsDesc, nil, Tex);
  DIRECTX.Device.CreateShaderResourceView(Tex, nil, TexSv);
  TexRtv := nil;
  if rendertarget then
    DIRECTX.Device.CreateRenderTargetView(Tex, nil, TexRtv);
end;

constructor TTexture.Create(sizeW, sizeH: Integer; rendertarget: Boolean;
  mipLevels: Integer = 1; sampleCount: Integer = 1);
begin
  inherited Create;

  Init(sizeW, sizeH, rendertarget, mipLevels, sampleCount);
end;

constructor TTexture.Create(rendertarget: Boolean; sizeW, sizeH: Integer;
  autoFillData: Integer = 0; sampleCount: Integer = 1);
begin
  if autoFillData <> 0 then
    Init(sizeW, sizeH, rendertarget, 8, sampleCount)
  else
    Init(sizeW, sizeH, rendertarget, 1, sampleCount);
  if autoFillData <> 0 then
    AutoFillTexture(autoFillData);
end;

procedure TTexture.FillTexture(pix: PDWORD);
var
  sizeW, sizeH: Integer;
  i, j, level: NativeInt;
  psrc, pdest: PByte;
begin
  //Make local ones, because will be reducing them
  sizeW := Self.SizeW;
  sizeH := Self.SizeH;
  for level := 0 to MipLevels-1 do
  begin
    DIRECTX.Context.UpdateSubresource(Tex, level, nil, Pointer(pix), sizeW * 4, sizeH * sizeW * 4);

    j := 0;
    while j < (sizeH and not 1) do
    begin
      psrc := PByte(pix) + (sizeW * j * 4);
      pdest := PByte(pix) + (sizeW * j);
      i := 0;
      while i < (sizeW shr 1) do
      begin
        pdest[0] := (psrc[0] + psrc[4] + psrc[sizeW * 4 + 0] + psrc[sizeW * 4 + 4]) shr 2;
        pdest[1] := (psrc[1] + psrc[5] + psrc[sizeW * 4 + 1] + psrc[sizeW * 4 + 5]) shr 2;
        pdest[2] := (psrc[2] + psrc[6] + psrc[sizeW * 4 + 2] + psrc[sizeW * 4 + 6]) shr 2;
        pdest[3] := (psrc[3] + psrc[7] + psrc[sizeW * 4 + 3] + psrc[sizeW * 4 + 7]) shr 2;
        Inc(i);
        Inc(psrc, 8);
        Inc(pdest, 4);
      end;
      Inc(j, 2);
    end;
    sizeW := sizeW shr 1;
    sizeH := sizeH shr 1;
  end;
end;

class procedure TTexture.ConvertToSRGB(linear: PDWORD);
var
  drgb: array[0..2] of DWORD;
  k: NativeInt;
  rgb: Single;
begin
  for k := Low(drgb) to High(drgb) do
  begin
    rgb := ((linear^ shr (k * 8)) and $FF) / 255.0;
    rgb := Power(rgb, 2.2);
    drgb[k] := Trunc(rgb * 255.0);
  end;
  linear^ := (linear^ and $FF000000) + (drgb[2] shl 16) + (drgb[1] shl 8) + (drgb[0] shl 0);
end;

procedure TTexture.AutoFillTexture(autoFillData: Integer);
var
  buf: array of DWORD;
  pix: PDWORDArray;
  i, j: NativeInt;
  curr: PDWORD;
begin
  SetLength(buf, SizeW * SizeH);
  pix := @buf[0];
  for j := 0 to SizeH-1 do
  begin
    for i := 0 to SizeW-1 do
    begin
      curr := @pix^[j*SizeW + i];
      case autoFillData of
        AUTO_WALL:
        begin
          if ((j div 4 and 15) = 0) or (((i div 4 and 15) = 0) and ((((i div 4 and 31) = 0) xor ((j div 4 shr 4) and 1 <> 0)) = False)) then
            curr^ := $FF3C3C3C
          else
            curr^ := $FFB4B4B4;
        end;
        AUTO_FLOOR:
        begin
          if ((i shr 7) xor (j shr 7)) and 1 <> 0 then
            curr^ := $FFB4B4B4
          else
            curr^ := $FF505050;
        end;
        AUTO_CEILING:
        begin
          if (i div 4 = 0) or (j div 4 = 0) then
            curr^ := $FF505050
          else
            curr^ := $FFB4B4B4;
        end;
        AUTO_WHITE:
        begin
          curr^ := $FFFFFFFF;
        end;
        AUTO_GRADE_256:
        begin
          curr^ := $FF000000 + i*$010101;
        end;
        AUTO_GRID:
        begin
          if (i<4) or (i>(SizeW - 5)) or (j<4) or (j>(SizeH - 5)) then
            curr^ := $FFFFFFFF
          else
            curr^ := $FF000000;
        end;
      else
        curr^ := $FFFFFFFF;
      end;
    end;
  end;
  FillTexture(PDWORD(pix));
end;

{ TMaterial }

constructor TMaterial.Create(t: TTexture; flags: DWORD = DEFAULT_FLAGS;
  vertexDesc: PD3D11_INPUT_ELEMENT_DESC = nil; numVertexDesc: Integer = 3;
  vertexShader: PAnsiChar = nil; pixelShader: PAnsiChar = nil; vSize: Integer = 24);
const
  defaultVertexDesc: array[0..2] of D3D11_INPUT_ELEMENT_DESC = (
    (SemanticName: 'Position'; SemanticIndex: 0; Format: DXGI_FORMAT_R32G32B32_FLOAT; InputSlot: 0; AlignedByteOffset: 0;  InputSlotClass: D3D11_INPUT_PER_VERTEX_DATA; InstanceDataStepRate: 0),
    (SemanticName: 'Color';    SemanticIndex: 0; Format: DXGI_FORMAT_B8G8R8A8_UNORM;  InputSlot: 0; AlignedByteOffset: 12; InputSlotClass: D3D11_INPUT_PER_VERTEX_DATA; InstanceDataStepRate: 0),
    (SemanticName: 'TexCoord'; SemanticIndex: 0; Format: DXGI_FORMAT_R32G32_FLOAT;    InputSlot: 0; AlignedByteOffset: 16; InputSlotClass: D3D11_INPUT_PER_VERTEX_DATA; InstanceDataStepRate: 0)
  );
  defaultVertexShaderSrc: PAnsiChar =
    'float4x4 ProjView;  float4 MasterCol;' +
    'void main(in  float4 Position  : POSITION,    in  float4 Color : COLOR0, in  float2 TexCoord  : TEXCOORD0,' +
    '          out float4 oPosition : SV_Position, out float4 oColor: COLOR0, out float2 oTexCoord : TEXCOORD0)' +
    '{   oPosition = mul(ProjView, Position); oTexCoord = TexCoord;' +
    '    oColor = MasterCol * Color; }';
  defaultPixelShaderSrc: PAnsiChar =
    'Texture2D Texture   : register(t0); SamplerState Linear : register(s0);' +
    'float4 main(in float4 Position : SV_Position, in float4 Color: COLOR0, in float2 TexCoord : TEXCOORD0) : SV_Target' +
    '{   float4 TexCol = Texture.Sample(Linear, TexCoord);' +
    '    if (TexCol.a==0) clip(-1);' + // If alpha = 0, don't draw
    '    return(Color * TexCol); }';
  // vertex shader for instanced stereo
  instancedStereoVertexShaderSrc: PAnsiChar =
    'float4x4 modelViewProj[2];  float4 MasterCol;' +
    'void main(in  float4 Position  : POSITION,    in  float4 Color : COLOR0, in  float2 TexCoord  : TEXCOORD0,' +
    '          in  uint instanceID : SV_InstanceID, ' +
    '          out float4 oPosition : SV_Position, out float4 oColor: COLOR0, out float2 oTexCoord : TEXCOORD0,' +
    '          out float oClipDist : SV_ClipDistance0, out float oCullDist : SV_CullDistance0)' +
    '{' +
    '   const float4 EyeClipPlane[2] = { { -1, 0, 0, 0 }, { 1, 0, 0, 0 } };' +
    '   uint eyeIndex = instanceID & 1;' +
    // transform to clip space for correct eye (includes offset and scale)
    '   oPosition = mul(modelViewProj[eyeIndex], Position); ' +
    // calculate distance from left/right clip plane (try setting to 0 to see why clipping is necessary)
    '   oCullDist = oClipDist = dot(EyeClipPlane[eyeIndex], oPosition);' +
    '   oTexCoord = TexCoord; ' +
    '   oColor = MasterCol * Color;' +
    '}';
var
  blobData, errorBlob: ID3DBlob;
  ss: D3D11_SAMPLER_DESC;
  rs: D3D11_RASTERIZER_DESC;
  dss: D3D11_DEPTH_STENCIL_DESC;
  bm: D3D11_BLEND_DESC;
  hr: HRESULT;
begin
  inherited Create;

  Tex := t;
  VertexSize := vSize;

  // Use defaults if no shaders specified
  if vertexDesc = nil then
  begin
    vertexDesc := @defaultVertexDesc;
    numVertexDesc := 3;
  end;
  if vertexShader = nil then vertexShader := defaultVertexShaderSrc;
  if pixelShader = nil then pixelShader := defaultPixelShaderSrc;

  // Create vertex shader
  hr := D3DCompile(vertexShader, Length(vertexShader), nil, nil, nil, 'main', 'vs_4_0', 0, 0, blobData, errorBlob);
  if Failed(hr) then
  begin
    MessageBoxA(0, PAnsiChar(errorBlob.GetBufferPointer), 'Error compiling vertex shader', MB_OK);
    ExitCode := -1;
    Halt;
  end;
  DIRECTX.Device.CreateVertexShader(blobData.GetBufferPointer, blobData.GetBufferSize, nil, @Self.VertexShader);

  // Create input layout
  hr := DIRECTX.Device.CreateInputLayout(vertexDesc, numVertexDesc,
    blobData.GetBufferPointer, blobData.GetBufferSize, InputLayout);
  blobData := nil;

  // Create vertex shader for instancing
  hr := D3DCompile(instancedStereoVertexShaderSrc, Length(instancedStereoVertexShaderSrc), nil, nil, nil, 'main', 'vs_4_0', 0, 0, blobData, errorBlob);
  if Failed(hr) then
  begin
    MessageBoxA(0, PAnsiChar(errorBlob.GetBufferPointer), 'Error compiling vertex shader', MB_OK);
    ExitCode := -1;
    Halt;
  end;
  DIRECTX.Device.CreateVertexShader(blobData.GetBufferPointer, blobData.GetBufferSize, nil, @VertexShaderInstanced);
  blobData := nil;

  // Create pixel shader
  hr := D3DCompile(pixelShader, Length(pixelShader), nil, nil, nil, 'main', 'ps_4_0', 0, 0, blobData, ID3D10Blob(nil^));
  VALIDATE(Succeeded(hr), 'Failed to compile pixel shader.');
  hr := DIRECTX.Device.CreatePixelShader(blobData.GetBufferPointer, blobData.GetBufferSize, nil, Self.PixelShader);
  blobData := nil;

  // Create sampler state
  FillChar(ss, SizeOf(ss), 0);
  if flags and Ord(MAT_WRAP) <> 0 then
    ss.AddressU := D3D11_TEXTURE_ADDRESS_WRAP
  else
    ss.AddressU := D3D11_TEXTURE_ADDRESS_BORDER;
  ss.AddressV := ss.AddressU;
  ss.AddressW := ss.AddressU;
  ss.Filter := D3D11_FILTER_ANISOTROPIC;
  ss.MaxAnisotropy := 8;
  ss.MaxLOD := 15;
  DIRECTX.Device.CreateSamplerState(ss, SamplerState);

  // Create rasterizer
  FillChar(rs, SizeOf(rs), 0);
  rs.AntialiasedLineEnable := True;
  rs.DepthClipEnable := True;
  if flags and Ord(MAT_NOCULL) <> 0 then
    rs.CullMode := D3D11_CULL_NONE
  else
    rs.CullMode := D3D11_CULL_BACK;
  if flags and Ord(MAT_WIRE) <> 0 then
    rs.FillMode := D3D11_FILL_WIREFRAME
  else
    rs.FillMode := D3D11_FILL_SOLID;
  DIRECTX.Device.CreateRasterizerState(rs, Rasterizer);

  // Create depth state
  FillChar(dss, SizeOf(dss), 0);
  dss.DepthEnable := True;
  if flags and Ord(MAT_ZALWAYS) <> 0 then
    dss.DepthFunc := D3D11_COMPARISON_ALWAYS
  else
    dss.DepthFunc := D3D11_COMPARISON_LESS;
  dss.DepthWriteMask := D3D11_DEPTH_WRITE_MASK_ALL;
  DIRECTX.Device.CreateDepthStencilState(dss, DepthState);

  //Create blend state - trans or otherwise
  FillChar(bm, SizeOf(bm), 0);
  bm.RenderTarget[0].BlendEnable := flags and Ord(MAT_TRANS) <> 0;
  bm.RenderTarget[0].BlendOp := D3D11_BLEND_OP_ADD;
  bm.RenderTarget[0].BlendOpAlpha := D3D11_BLEND_OP_ADD;
  bm.RenderTarget[0].SrcBlend := D3D11_BLEND_SRC_ALPHA;
  bm.RenderTarget[0].SrcBlendAlpha := D3D11_BLEND_SRC_ALPHA;
  bm.RenderTarget[0].DestBlend := D3D11_BLEND_INV_SRC_ALPHA;
  bm.RenderTarget[0].DestBlendAlpha := D3D11_BLEND_INV_SRC_ALPHA;
  bm.RenderTarget[0].RenderTargetWriteMask := Ord(D3D11_COLOR_WRITE_ENABLE_ALL);
  DIRECTX.Device.CreateBlendState(bm, BlendState);
end;

destructor TMaterial.Destroy;
begin
  FreeAndNil(Tex);

  inherited;
end;

{ TVertex }

class function TVertex.Create(pos: TPoint3D; c: DWORD; u, v: Single): TVertex;
begin
  Result.Pos := pos;
  Result.C := c;
  Result.U := u;
  Result.V := v;
end;

{ TTriangleSet }

procedure TTriangleSet.AddQuad(const v0, v1, v2, v3: TVertex);
begin
  AddTriangle(v0, v1, v2);
  AddTriangle(v3, v2, v1);
end;

procedure TTriangleSet.AddTriangle(const v0, v1, v2: TVertex);
begin
  FIndices := FIndices + [NumVertices+0, NumVertices+1, NumVertices+2];
  FVertices := FVertices + [v0, v1, v2];
end;

function TTriangleSet.ModifyColor(c: DWORD; pos: TPoint3D): DWORD;

  function GetLengthLocal(x, y, z: Single): Single;
  begin
    Result := Sqrt(x*x + y*y + z*z);
  end;

var
  dist1, dist2, dist3: Single;
  bri: Integer;
  R, G, B: Single;
begin
  dist1 := GetLengthLocal(pos.x - (-2), pos.y - (4), pos.z - (-2));
  dist2 := GetLengthLocal(pos.x - (3),  pos.y - (4), pos.z - (-3));
  dist3 := GetLengthLocal(pos.x - (-4), pos.y - (3), pos.z - (25));
  bri := Random(160);
  R := ((c shr 16) and $FF) * (bri + 192.0*(0.65 + 8 / dist1 + 1 / dist2 + 4 / dist3)) / 255.0;
  G := ((c shr 8) and $FF) * (bri + 192.0*(0.65 + 8 / dist1 + 1 / dist2 + 4 / dist3)) / 255.0;
  B := ((c shr 0) and $FF) * (bri + 192.0*(0.65 + 8 / dist1 + 1 / dist2 + 4 / dist3)) / 255.0;
  Result := (c and $FF000000) + (IfThen(R>255, 255, Trunc(R)) shl 16) + (IfThen(G>255, 255, Trunc(G)) shl 8) + IfThen(B>255, 255, Trunc(B));
end;

procedure TTriangleSet.AddSolidColorBox(x1, y1, z1, x2, y2, z2: Single; c: DWORD);
begin
  AddQuad(TVertex.Create(Point3D(x1, y2, z1), ModifyColor(c, Point3D(x1, y2, z1)), z1, x1),
          TVertex.Create(Point3D(x2, y2, z1), ModifyColor(c, Point3D(x2, y2, z1)), z1, x2),
          TVertex.Create(Point3D(x1, y2, z2), ModifyColor(c, Point3D(x1, y2, z2)), z2, x1),
          TVertex.Create(Point3D(x2, y2, z2), ModifyColor(c, Point3D(x2, y2, z2)), z2, x2));
  AddQuad(TVertex.Create(Point3D(x2, y1, z1), ModifyColor(c, Point3D(x2, y1, z1)), z1, x2),
          TVertex.Create(Point3D(x1, y1, z1), ModifyColor(c, Point3D(x1, y1, z1)), z1, x1),
          TVertex.Create(Point3D(x2, y1, z2), ModifyColor(c, Point3D(x2, y1, z2)), z2, x2),
          TVertex.Create(Point3D(x1, y1, z2), ModifyColor(c, Point3D(x1, y1, z2)), z2, x1));
  AddQuad(TVertex.Create(Point3D(x1, y1, z2), ModifyColor(c, Point3D(x1, y1, z2)), z2, y1),
          TVertex.Create(Point3D(x1, y1, z1), ModifyColor(c, Point3D(x1, y1, z1)), z1, y1),
          TVertex.Create(Point3D(x1, y2, z2), ModifyColor(c, Point3D(x1, y2, z2)), z2, y2),
          TVertex.Create(Point3D(x1, y2, z1), ModifyColor(c, Point3D(x1, y2, z1)), z1, y2));
  AddQuad(TVertex.Create(Point3D(x2, y1, z1), ModifyColor(c, Point3D(x2, y1, z1)), z1, y1),
          TVertex.Create(Point3D(x2, y1, z2), ModifyColor(c, Point3D(x2, y1, z2)), z2, y1),
          TVertex.Create(Point3D(x2, y2, z1), ModifyColor(c, Point3D(x2, y2, z1)), z1, y2),
          TVertex.Create(Point3D(x2, y2, z2), ModifyColor(c, Point3D(x2, y2, z2)), z2, y2));
  AddQuad(TVertex.Create(Point3D(x1, y1, z1), ModifyColor(c, Point3D(x1, y1, z1)), x1, y1),
          TVertex.Create(Point3D(x2, y1, z1), ModifyColor(c, Point3D(x2, y1, z1)), x2, y1),
          TVertex.Create(Point3D(x1, y2, z1), ModifyColor(c, Point3D(x1, y2, z1)), x1, y2),
          TVertex.Create(Point3D(x2, y2, z1), ModifyColor(c, Point3D(x2, y2, z1)), x2, y2));
  AddQuad(TVertex.Create(Point3D(x2, y1, z2), ModifyColor(c, Point3D(x2, y1, z2)), x2, y1),
          TVertex.Create(Point3D(x1, y1, z2), ModifyColor(c, Point3D(x1, y1, z2)), x1, y1),
          TVertex.Create(Point3D(x2, y2, z2), ModifyColor(c, Point3D(x2, y2, z2)), x2, y2),
          TVertex.Create(Point3D(x1, y2, z2), ModifyColor(c, Point3D(x1, y2, z2)), x1, y2));
end;

function TTriangleSet.GetNumVertices: Integer;
begin
  Result := Length(FVertices);
end;

function TTriangleSet.GetNumIndices: Integer;
begin
  Result := Length(FIndices);
end;

{ TRenderItem }

procedure TRenderItem.Update;
begin
  { Do nothing }
end;

{ TModel }

procedure TModel.Init(const t: TTriangleSet);
begin
  FNumIndices := t.NumIndices;
  FVertexBuffer := TDataBuffer.Create(DIRECTX.Device, D3D11_BIND_VERTEX_BUFFER, @t.Vertices[0], t.NumVertices * SizeOf(TVertex));
  FIndexBuffer := TDataBuffer.Create(DIRECTX.Device, D3D11_BIND_INDEX_BUFFER, @t.Indices[0], t.NumIndices * SizeOf(SmallInt));
end;

constructor TModel.Create(const t: TTriangleSet; argPos: TPoint3D; argRot: TQuaternion3D; argFill: TMaterial);
begin
  inherited Create;

  FVisible := True;
  FPos := argPos;
  FRot := argRot;
  FFill := argFill;
  Init(t);
end;

// 2D scenes, for latency tester and full screen copies, etc
constructor TModel.Create(mat: TMaterial; minx, miny, maxx, maxy: Single; zDepth: Single = 0);
var
  quad: TTriangleSet;
begin
  inherited Create;

  FVisible := True;
  FPos := Point3D(0, 0, 0);
  FRot := TQuaternion3D.Identity;
  FFill := mat;

  quad := Default(TTriangleSet);
  quad.AddQuad(TVertex.Create(Point3D(minx, miny, zDepth), $FFFFFFFF, 0, 1),
               TVertex.Create(Point3D(minx, maxy, zDepth), $FFFFFFFF, 0, 0),
               TVertex.Create(Point3D(maxx, miny, zDepth), $FFFFFFFF, 1, 1),
               TVertex.Create(Point3D(maxx, maxy, zDepth), $FFFFFFFF, 1, 0));
  Init(quad);
end;

destructor TModel.Destroy;
begin
  FreeAndNil(FFill);
  FreeAndNil(FVertexBuffer);
  FreeAndNil(FIndexBuffer);

  inherited;
end;

procedure TModel.Render(const projView: TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean);
var
  modelMat, mat: TMatrix3D;
  col: array[0..3] of Single;
  map: D3D11_MAPPED_SUBRESOURCE;
  offset: UINT;
begin
  if not FVisible then Exit;

  modelMat := TMatrix3D(FRot) * TMatrix3D.CreateTranslation(FPos);
  mat := modelMat * projView;
  col[0] := R;
  col[1] := G;
  col[2] := B;
  col[3] := A;
  if standardUniforms then
  begin
    Move(mat, DIRECTX.UniformData[0], 64); // ProjView
    Move(col, DIRECTX.UniformData[64], 16); // MasterCol
  end;
  DIRECTX.Context.Map(DIRECTX.UniformBufferGen.D3DBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, map);
  Move(DIRECTX.UniformData, map.pData^, DIRECTX.UNIFORM_DATA_SIZE);
  DIRECTX.Context.Unmap(DIRECTX.UniformBufferGen.D3DBuffer, 0);
  DIRECTX.Context.IASetInputLayout(FFill.InputLayout);
  DIRECTX.Context.IASetIndexBuffer(FIndexBuffer.D3DBuffer, DXGI_FORMAT_R16_UINT, 0);
  offset := 0;
  DIRECTX.Context.IASetVertexBuffers(0, 1, FVertexBuffer.D3DBuffer, @FFill.VertexSize, @offset);
  DIRECTX.Context.IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
  DIRECTX.Context.VSSetShader(FFill.VertexShader, nil, 0);
  DIRECTX.Context.PSSetShader(FFill.PixelShader, nil, 0);
  DIRECTX.Context.PSSetSamplers(0, 1, FFill.SamplerState);
  DIRECTX.Context.RSSetState(FFill.Rasterizer);
  DIRECTX.Context.OMSetDepthStencilState(FFill.DepthState, 0);
  DIRECTX.Context.OMSetBlendState(FFill.BlendState, TFourSingleArray(nil^), $FFFFFFFF);
  DIRECTX.Context.PSSetShaderResources(0, 1, FFill.Tex.TexSv);
  DIRECTX.Context.DrawIndexed(UINT(FNumIndices), 0, 0);
end;

// render using stereo instancing
procedure TModel.RenderInstanced(const viewProjMats: array of TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean);
var
  modelMat, modelViewProjMatL, modelViewProjMatR: TMatrix3D;
  col: array[0..3] of Single;
  map: D3D11_MAPPED_SUBRESOURCE;
  offset: UINT;
begin
  if not FVisible then Exit;

  modelMat := TMatrix3D(FRot) * TMatrix3D.CreateTranslation(FPos);
  modelViewProjMatL := modelMat * viewProjMats[0];
  modelViewProjMatR := modelMat * viewProjMats[1];
  col[0] := R;
  col[1] := G;
  col[2] := B;
  col[3] := A;
  if standardUniforms then
  begin
    Move(modelViewProjMatL, DIRECTX.UniformData[0], 64);
    Move(modelViewProjMatR, DIRECTX.UniformData[64], 64);
    Move(col, DIRECTX.UniformData[128], 16); // MasterCol
  end;
  DIRECTX.Context.Map(DIRECTX.UniformBufferGen.D3DBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, map);
  Move(DIRECTX.UniformData, map.pData^, DIRECTX.UNIFORM_DATA_SIZE);
  DIRECTX.Context.Unmap(DIRECTX.UniformBufferGen.D3DBuffer, 0);
  DIRECTX.Context.IASetInputLayout(FFill.InputLayout);
  DIRECTX.Context.IASetIndexBuffer(FIndexBuffer.D3DBuffer, DXGI_FORMAT_R16_UINT, 0);
  offset := 0;
  DIRECTX.Context.IASetVertexBuffers(0, 1, FVertexBuffer.D3DBuffer, @FFill.VertexSize, @offset);
  DIRECTX.Context.IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
  DIRECTX.Context.VSSetShader(FFill.VertexShaderInstanced, nil, 0);
  DIRECTX.Context.PSSetShader(FFill.PixelShader, nil, 0);
  DIRECTX.Context.PSSetSamplers(0, 1, FFill.SamplerState);
  DIRECTX.Context.RSSetState(FFill.Rasterizer);
  DIRECTX.Context.OMSetDepthStencilState(FFill.DepthState, 0);
  DIRECTX.Context.OMSetBlendState(FFill.BlendState, TFourSingleArray(nil^), $FFFFFFFF);
  DIRECTX.Context.PSSetShaderResources(0, 1, FFill.Tex.TexSv);

  // draw 2 instances
  DIRECTX.Context.DrawIndexedInstanced(UINT(FNumIndices), 2, 0, 0, 0);
end;

{ TScene }

constructor TScene.Create(includeIntensiveGPUobject: Boolean);
begin
  inherited Create;

  FModels := TObjectList<TRenderItem>.Create;
  Init(includeIntensiveGPUobject);
end;

destructor TScene.Destroy;
begin
  FModels.Free;

  inherited;
end;

procedure TScene.Add(n: TRenderItem);
begin
  FModels.Add(n);
end;

procedure TScene.Render(const projView: TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean);
var
  i: NativeInt;
begin
  for i := 0 to FModels.Count-1 do
    FModels[i].Render(projView, R, G, B, A, standardUniforms);
end;

procedure TScene.RenderInstanced(const projViews: array of TMatrix3D; R, G, B, A: Single; standardUniforms: Boolean);
var
  i: NativeInt;
begin
  for i := 0 to FModels.Count-1 do
    Models[i].RenderInstanced(projViews, R, G, B, A, standardUniforms);
end;

procedure TScene.Init(includeIntensiveGPUobject: Boolean);
var
  cube: TTriangleSet;
  spareCube: TTriangleSet;
  walls: TTriangleSet;
  partitions: TTriangleSet;
  depth: Single;
  floors: TTriangleSet;
  ceiling: TTriangleSet;
  furniture: TTriangleSet;
  f: Single;
begin
  cube := Default(TTriangleSet);
  cube.AddSolidColorBox(0.5, -0.5, 0.5, -0.5, 0.5, -0.5, $FF404040);
  Add(
    TModel.Create(cube, Point3D(0, 0, 0), TQuaternion3D.Identity,
      TMaterial.Create(
        TTexture.Create(False, 256, 256, TTexture.AUTO_CEILING)
      )
    )
  );

  spareCube := Default(TTriangleSet);
  spareCube.AddSolidColorBox(0.1, -0.1, 0.1, -0.1, +0.1, -0.1, $FFFF0000);
  Add(
    TModel.Create(spareCube, Point3D(0, -10, 0), TQuaternion3D.Identity,
      TMaterial.Create(
        TTexture.Create(False, 256, 256, TTexture.AUTO_CEILING)
      )
    )
  );

  walls := Default(TTriangleSet);
  walls.AddSolidColorBox(10.1, 0.0, 20.0, 10.0, 4.0, -20.0, $FF808080);    // Left Wall
  walls.AddSolidColorBox(10.0, -0.1, 20.1, -10.0, 4.0, 20.0, $FF808080);   // Back Wall
  walls.AddSolidColorBox(-10.0, -0.1, 20.0, -10.1, 4.0, -20.0, $FF808080); // Right Wall
  Add(
    TModel.Create(walls, Point3D(0, 0, 0), TQuaternion3D.Identity,
      TMaterial.Create(
        TTexture.Create(false, 256, 256, TTexture.AUTO_WALL)
      )
    )
  );

  if includeIntensiveGPUobject then
  begin
    partitions := Default(TTriangleSet);
    depth := 0.0;
    while depth > -3.0 do
    begin
      partitions.AddSolidColorBox(9.0, 0.5, -depth, -9.0, 3.5, -depth, $10FF80FF); // Partition
      depth := depth - 0.1;
    end;
    Add(
      TModel.Create(partitions, Point3D(0, 0, 0), TQuaternion3D.Identity,
        TMaterial.Create(
          TTexture.Create(False, 256, 256, TTexture.AUTO_FLOOR)
        )
      )
    ); // Floors
  end;

  floors := Default(TTriangleSet);
  floors.AddSolidColorBox(10.0, -0.1, 20.0, -10.0, 0.0, -20.1, $FF808080); // Main floor
  floors.AddSolidColorBox(15.0, -6.1, -18.0, -15.0, -6.0, -30.0, $FF808080); // Bottom floor
  Add(
    TModel.Create(floors, Point3D(0, 0, 0), TQuaternion3D.Identity,
      TMaterial.Create(
        TTexture.Create(False, 256, 256, TTexture.AUTO_FLOOR)
      )
    )
  ); // Floors

  ceiling := Default(TTriangleSet);
  ceiling.AddSolidColorBox(10.0, 4.0, 20.0, -10.0, 4.1, -20.1, $FF808080);
  Add(
    TModel.Create(ceiling, Point3D(0, 0, 0), TQuaternion3D.Identity,
      TMaterial.Create(
        TTexture.Create(False, 256, 256, TTexture.AUTO_CEILING)
      )
    )
  ); // Ceiling

  furniture := Default(TTriangleSet);
  furniture.AddSolidColorBox(-9.5, 0.75, -3.0, -10.1, 2.5, -3.1, $FF383838);    // Right side shelf// Verticals
  furniture.AddSolidColorBox(-9.5, 0.95, -3.7, -10.1, 2.75, -3.8, $FF383838);   // Right side shelf
  furniture.AddSolidColorBox(-9.55, 1.20, -2.5, -10.1, 1.30, -3.75, $FF383838); // Right side shelf// Horizontals
  furniture.AddSolidColorBox(-9.55, 2.00, -3.05, -10.1, 2.10, -4.2, $FF383838); // Right side shelf
  furniture.AddSolidColorBox(-5.0, 1.1, -20.0, -10.0, 1.2, -20.1, $FF383838);   // Right railing
  furniture.AddSolidColorBox(10.0, 1.1, -20.0, 5.0, 1.2, -20.1, $FF383838);   // Left railing
  f := 5;
  while f <= 9 do
  begin
    furniture.AddSolidColorBox(-f, 0.0, -20.0, -f - 0.1, 1.1, -20.1, $FF505050); // Left Bars
    f := f + 1;
  end;
  f := 5;
  while f <= 9 do
  begin
    furniture.AddSolidColorBox(f, 1.1, -20.0, f + 0.1, 0.0, -20.1, $FF505050); // Right Bars
    f := f + 1;
  end;
  furniture.AddSolidColorBox(1.8, 0.8, -1.0, 0.0, 0.7, 0.0, $FF505000);  // Table
  furniture.AddSolidColorBox(1.8, 0.0, 0.0, 1.7, 0.7, -0.1, $FF505000); // Table Leg
  furniture.AddSolidColorBox(1.8, 0.7, -1.0, 1.7, 0.0, -0.9, $FF505000); // Table Leg
  furniture.AddSolidColorBox(0.0, 0.0, -1.0, 0.1, 0.7, -0.9, $FF505000);  // Table Leg
  furniture.AddSolidColorBox(0.0, 0.7, 0.0, 0.1, 0.0, -0.1, $FF505000);  // Table Leg
  furniture.AddSolidColorBox(1.4, 0.5, 1.1, 0.8, 0.55, 0.5, $FF202050);  // Chair Set
  furniture.AddSolidColorBox(1.401, 0.0, 1.101, 1.339, 1.0, 1.039, $FF202050); // Chair Leg 1
  furniture.AddSolidColorBox(1.401, 0.5, 0.499, 1.339, 0.0, 0.561, $FF202050); // Chair Leg 2
  furniture.AddSolidColorBox(0.799, 0.0, 0.499, 0.861, 0.5, 0.561, $FF202050); // Chair Leg 2
  furniture.AddSolidColorBox(0.799, 1.0, 1.101, 0.861, 0.0, 1.039, $FF202050); // Chair Leg 2
  furniture.AddSolidColorBox(1.4, 0.97, 1.05, 0.8, 0.92, 1.10, $FF202050); // Chair Back high bar
  f := 3.0;;
  while f <= 6.6 do
  begin
    furniture.AddSolidColorBox(3, 0.0, -f, 2.9, 1.3, -f - 0.1, $FF404040); // Posts
    f := f + 0.4;
  end;
  Add(
    TModel.Create(furniture, Point3D(0, 0, 0), TQuaternion3D.Identity,
      TMaterial.Create(
        TTexture.Create(False, 256, 256, TTexture.AUTO_WHITE)
      )
    )
  ); // Fixtures & furniture
end;

procedure TScene.Update;
var
  i: NativeInt;
begin
  for i := 0 to FModels.Count-1 do
    FModels[i].Update;
end;

{ TCamera }

constructor TCamera.Create(const APos: TPoint3D; const ARot: TQuaternion3D);
begin
  inherited Create;

  FPos := APos;
  FRot := ARot;
end;

function TCamera.GetViewMatrix: TMatrix3D;
var
  forward_: TPoint3D;
begin
  forward_ := Point3D(0, 0, -1) * TMatrix3D(FRot);
  Result := TMatrix3D.CreateLookAtRH(FPos, FPos + forward_, Point3D(0, 1, 0) * TMatrix3D(FRot));
end;

initialization
  DIRECTX := TDirectX11.Create;
finalization
  DIRECTX.Free;
end.
