unit UMathHelper;

interface

uses
  System.SysUtils, System.Math, System.Math.Vectors;

type
  TPoint3DHelper = record helper for TPoint3D
  public
    function Lerp(const APoint: TPoint3D; const ARate: Single): TPoint3D;
  end;

  TQuaternion3DHelper = record helper for TQuaternion3D
  public
  const
    Identity: TQuaternion3D = (ImagPart: (X: 0; Y: 0; Z: 0); RealPart: 1);
    function DotProduct(const AQuaternion: TQuaternion3D): Double;
    class function Slerp(const AQuaternion1, AQuaternion2: TQuaternion3D; const ARate: Single): TQuaternion3D; static;
//    class operator Negative(const AQuaternion: TQuaternion3D): TQuaternion3D;
    function Negative: TQuaternion3D;
//    class operator Add(const AQuaternion1, AQuaternion2: TQuaternion3D): TQuaternion3D;
    class function Add(const AQuaternion1, AQuaternion2: TQuaternion3D): TQuaternion3D; static;
//    class operator Divide(const AQuaternion: TQuaternion3D;
//      const AValue: Single): TQuaternion3D;
    function Divide(const AValue: Single): TQuaternion3D;
    function Scale(const AValue: Single): TQuaternion3D;
  end;

implementation

function MixValue(const V1, V2, Rate: Single): Single;
begin
  Result := V1 + Rate * (V2 - V1);
end;

{ TPoint3DHelper }

function TPoint3DHelper.Lerp(const APoint: TPoint3D;
  const ARate: Single): TPoint3D;
begin
  Result.X := Self.X + (APoint.X - Self.X) * ARate;
  Result.Y := Self.Y + (APoint.Y - Self.Y) * ARate;
  Result.Z := Self.Z + (APoint.Z - Self.Z) * ARate;
end;

{ TQuaternion3DHelper }

function TQuaternion3DHelper.DotProduct(const AQuaternion: TQuaternion3D): Double;
begin
  Result := Self.ImagPart.X * AQuaternion.ImagPart.X
    + Self.ImagPart.Y * AQuaternion.ImagPart.Y
    + Self.ImagPart.Z * AQuaternion.ImagPart.Z
    + Self.RealPart * AQuaternion.RealPart;
end;

class function TQuaternion3DHelper.Slerp(const AQuaternion1, AQuaternion2: TQuaternion3D; const ARate: Single): TQuaternion3D;
var
  z: TQuaternion3D;
  cosTheta: Extended;
  angle: Extended;
begin
  cosTheta := AQuaternion1.DotProduct(AQuaternion2);
  if cosTheta < 0 then
  begin
    z := AQuaternion2.Negative;
    cosTheta := -cosTheta;
  end
  else begin
    z := AQuaternion2;
  end;

  if cosTheta > 1 - Single.Epsilon then
  begin
    Result.ImagPart.X := MixValue(AQuaternion1.ImagPart.X, z.ImagPart.X, ARate);
    Result.ImagPart.Y := MixValue(AQuaternion1.ImagPart.Y, z.ImagPart.Y, ARate);
    Result.ImagPart.Z := MixValue(AQuaternion1.ImagPart.Z, z.ImagPart.Z, ARate);
    Result.RealPart := MixValue(AQuaternion1.RealPart, z.RealPart, ARate);
  end
  else begin
    angle := ArcCos(cosTheta);
    Result := TQuaternion3D.Add(AQuaternion1.Scale(Sin((1 - ARate) * angle)), z.Scale(Sin(ARate * angle))).Divide(Sin(angle));
  end;
end;

{
class operator TQuaternion3DHelper.Negative(const AQuaternion: TQuaternion3D): TQuaternion3D;
begin
  Result.ImagPart.X := -AQuaternion.ImagPart.X;
  Result.ImagPart.Y := -AQuaternion.ImagPart.Y;
  Result.ImagPart.Z := -AQuaternion.ImagPart.Z;
  Result.RealPart := -AQuaternion.RealPart;
end;
}
function TQuaternion3DHelper.Negative: TQuaternion3D;
begin
  Result.ImagPart.X := -Self.ImagPart.X;
  Result.ImagPart.Y := -Self.ImagPart.Y;
  Result.ImagPart.Z := -Self.ImagPart.Z;
  Result.RealPart := -Self.RealPart;
end;

{
class operator TQuaternion3DHelper.Add(const AQuaternion1, AQuaternion2: TQuaternion3D): TQuaternion3D;
begin
  Result.ImagPart.X := AQuaternion1.ImagPart.X + AQuaternion2.ImagPart.X;
  Result.ImagPart.Y := AQuaternion1.ImagPart.Y + AQuaternion2.ImagPart.Y;
  Result.ImagPart.Z := AQuaternion1.ImagPart.Z + AQuaternion2.ImagPart.Z;
  Result.RealPart := AQuaternion1.RealPart + AQuaternion2.RealPart;
end;
}
class function TQuaternion3DHelper.Add(const AQuaternion1, AQuaternion2: TQuaternion3D): TQuaternion3D;
begin
  Result.ImagPart.X := AQuaternion1.ImagPart.X + AQuaternion2.ImagPart.X;
  Result.ImagPart.Y := AQuaternion1.ImagPart.Y + AQuaternion2.ImagPart.Y;
  Result.ImagPart.Z := AQuaternion1.ImagPart.Z + AQuaternion2.ImagPart.Z;
  Result.RealPart := AQuaternion1.RealPart + AQuaternion2.RealPart;
end;

{
class operator TQuaternion3DHelper.Divide(const AQuaternion: TQuaternion3D;
  const AValue: Single): TQuaternion3D;
begin
  if AValue = 0.0 then
    Exit(TQuaternion3D.Identity);
  Result.ImagPart.X := AQuaternion.ImagPart.X / AValue;
  Result.ImagPart.Y := AQuaternion.ImagPart.Y / AValue;
  Result.ImagPart.Z := AQuaternion.ImagPart.Z / AValue;
  Result.RealPart := AQuaternion.RealPart / AValue;
end;
}
function TQuaternion3DHelper.Divide(const AValue: Single): TQuaternion3D;
begin
  if AValue = 0.0 then
    Exit(TQuaternion3D.Identity);
  Result.ImagPart.X := Self.ImagPart.X / AValue;
  Result.ImagPart.Y := Self.ImagPart.Y / AValue;
  Result.ImagPart.Z := Self.ImagPart.Z / AValue;
  Result.RealPart := Self.RealPart / AValue;
end;

function TQuaternion3DHelper.Scale(const AValue: Single): TQuaternion3D;
begin
  Result.ImagPart.X := Self.ImagPart.X * AValue;
  Result.ImagPart.Y := Self.ImagPart.Y * AValue;
  Result.ImagPart.Z := Self.ImagPart.Z * AValue;
  Result.RealPart := Self.RealPart * AValue;
end;

end.
