(********************************************************************************)(**
\file      OVR_CAPI.pas
\brief     C Interface to the Oculus PC SDK tracking and rendering library.
\copyright Copyright 2014 Oculus VR, LLC All Rights reserved.
************************************************************************************)
unit OVR_CAPI;

interface

uses
  OVR_CAPI_Keys, OVR_Version, OVR_ErrorCode;


//-----------------------------------------------------------------------------------
// ***** Word Size
//
/// Specifies the size of a pointer on the given platform.
///
const
  OVR_PTR_SIZE = SizeOf(Pointer);


//-----------------------------------------------------------------------------------
// ***** ovrBool

type
  ovrBool = (     ///< Boolean type
    ovrFalse = 0, ///< ovrBool value of false.
    ovrTrue  = 1  ///< ovrBool value of true.
  );
  TOvrBool = ovrBool;


//-----------------------------------------------------------------------------------
// ***** Simple Math Structures

type
  /// A 2D vector with integer components.
  ovrVector2i = record
    x, y: Int32;
  end align 4;
  TOvrVector2i = ovrVector2i;

  /// A 2D size with integer components.
  ovrSizei = record
    w, h: Int32;
  end align 4;
  TOvrSizei = ovrSizei;

  /// A 2D rectangle with a position and size.
  /// All components are integers.
  ovrRecti = record
    Pos: TOvrVector2i;
    Size: TOvrSizei;
  end align 4;
  TOvrRecti = ovrRecti;

  /// A quaternion rotation.
  ovrQuatf = record
    x, y, z, w: Single;
  end align 4;
  TOvrQuatf = ovrQuatf;

  /// A 2D vector with float components.
  ovrVector2f = record
    x, y: Single;
  end align 4;
  TOvrVector2f = ovrVector2f;

  /// A 3D vector with float components.
  ovrVector3f = record
    x, y, z: Single;
  end align 4;
  TOvrVector3f = ovrVector3f;

  /// A 4x4 matrix with float elements.
  ovrMatrix4f = record
    M: array[0..3] of array[0..3] of Single;
  end align 4;
  TOvrMatrix4f = ovrMatrix4f;

  /// Position and orientation together.
  POvrPosef = ^TOvrPosef;
  ovrPosef = record
    Orientation: TOvrQuatf;
    Position: TOvrVector3f;
  end align 4;
  TOvrPosef = ovrPosef;

  /// A full pose (rigid body) configuration with first and second derivatives.
  ///
  /// Body refers to any object for which ovrPoseStatef is providing data.
  /// It can be the HMD, Touch controller, sensor or something else. The context 
  /// depends on the usage of the struct.
  ovrPoseStatef = record
    ThePose: TOvrPosef;                ///< Position and orientation.
    AngularVelocity: TOvrVector3f;     ///< Angular velocity in radians per second.
    LinearVelocity: TOvrVector3f;      ///< Velocity in meters per second.
    AngularAcceleration: TOvrVector3f; ///< Angular acceleration in radians per second per second.
    LinearAcceleration: TOvrVector3f;  ///< Acceleration in meters per second per second.
    pad0: array[0..3] of Byte;         ///< \internal struct pad.
    TimeInSeconds: Double;             ///< Absolute time that this pose refers to. \see ovr_GetTimeInSeconds
  end align 8;
  TOvrPoseStatef = ovrPoseStatef;

  /// Describes the up, down, left, and right angles of the field of view.
  ///
  /// Field Of View (FOV) tangent of the angle units.
  /// \note For a standard 90 degree vertical FOV, we would
  /// have: { UpTan = tan(90 degrees / 2), DownTan = tan(90 degrees / 2) }.
  POvrFovPort = ^TOvrFovPort;
  ovrFovPort = record
    UpTan: Single;    ///< The tangent of the angle between the viewing vector and the top edge of the field of view.
    DownTan: Single;  ///< The tangent of the angle between the viewing vector and the bottom edge of the field of view.
    LeftTan: Single;  ///< The tangent of the angle between the viewing vector and the left edge of the field of view.
    RightTan: Single; ///< The tangent of the angle between the viewing vector and the right edge of the field of view.
  end align 4;
  TOvrFovPort = ovrFovPort;

//-----------------------------------------------------------------------------------
// ***** HMD Types

  /// Enumerates all HMD types that we support.
  ///
  /// The currently released developer kits are ovrHmd_DK1 and ovrHmd_DK2. The other enumerations are for internal use only.
  ovrHmdType = (
    ovrHmd_None      = 0,
    ovrHmd_DK1       = 3,
    ovrHmd_DKHD      = 4,
    ovrHmd_DK2       = 6,
    ovrHmd_CB        = 8,
    ovrHmd_Other     = 9,
    ovrHmd_E3_2015   = 10,
    ovrHmd_ES06      = 11,
    ovrHmd_ES09      = 12,
    ovrHmd_ES11      = 13,
    ovrHmd_CV1       = 14,

    ovrHmd_EnumSize  = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrHmdType = ovrHmdType;

  /// HMD capability bits reported by device.
  ///
  ovrHmdCaps = (
    // Read-only flags
    ovrHmdCap_DebugDevice   = $0010,   ///< <B>(read only)</B> Specifies that the HMD is a virtual debug device.


    ovrHmdCap_EnumSize      = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrHmdCaps = ovrHmdCaps;

  /// Tracking capability bits reported by the device.
  /// Used with ovr_GetTrackingCaps.
  ovrTrackingCaps = (
    ovrTrackingCap_Orientation      = $0010,    ///< Supports orientation tracking (IMU).
    ovrTrackingCap_MagYawCorrection = $0020,    ///< Supports yaw drift correction via a magnetometer or other means.
    ovrTrackingCap_Position         = $0040,    ///< Supports positional tracking.
    ovrTrackingCap_EnumSize         = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrTrackingCaps = ovrTrackingCaps;

  /// Specifies which eye is being used for rendering.
  /// This type explicitly does not include a third "NoStereo" monoscopic option, as such is
  /// not required for an HMD-centered API.
  ovrEyeType = (
    ovrEye_Left     = 0,        ///< The left eye, from the viewer's perspective.
    ovrEye_Right    = 1,        ///< The right eye, from the viewer's perspective.
    ovrEye_Count    = 2,        ///< \internal Count of enumerated elements.
    ovrEye_EnumSize = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrEyeType = ovrEyeType;

  /// Specifies the coordinate system ovrTrackingState returns tracking poses in.
  /// Used with ovr_SetTrackingOriginType()
  ovrTrackingOrigin = (
    /// \brief Tracking system origin reported at eye (HMD) height
    /// \details Prefer using this origin when your application requires
    /// matching user's current physical head pose to a virtual head pose
    /// without any regards to a the height of the floor. Cockpit-based,
    /// or 3rd-person experiences are ideal candidates.
    /// When used, all poses in ovrTrackingState are reported as an offset
    /// transform from the profile calibrated or recentered HMD pose.
    /// It is recommended that apps using this origin type call ovr_RecenterTrackingOrigin
    /// prior to starting the VR experience, but notify the user before doing so
    /// to make sure the user is in a comfortable pose, facing a comfortable
    /// direction.
    ovrTrackingOrigin_EyeLevel = 0,
    /// \brief Tracking system origin reported at floor height
    /// \details Prefer using this origin when your application requires the
    /// physical floor height to match the virtual floor height, such as
    /// standing experiences.
    /// When used, all poses in ovrTrackingState are reported as an offset
    /// transform from the profile calibrated floor pose. Calling ovr_RecenterTrackingOrigin
    /// will recenter the X & Z axes as well as yaw, but the Y-axis (i.e. height) will continue
    /// to be reported using the floor height as the origin for all poses.
    ovrTrackingOrigin_FloorLevel = 1,
    ovrTrackingOrigin_Count = 2,           ///< \internal Count of enumerated elements.
    ovrTrackingOrigin_EnumSize = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrTrackingOrigin = ovrTrackingOrigin;

  /// Identifies a graphics device in a platform-specific way.
  /// For Windows this is a LUID type.
  ovrGraphicsLuid = record
    // Public definition reserves space for graphics API-specific implementation
    Reserved: array[0..7] of AnsiChar;
  end align OVR_PTR_SIZE;
  TOvrGraphicsLuid = ovrGraphicsLuid;

  /// This is a complete descriptor of the HMD.
  ovrHmdDesc = record
    Type_: TOvrHmdType;                                          ///< The type of HMD.
{$IFDEF CPU64BITS}
    pad0: array[0..3] of Byte;                                   ///< \internal struct paddding.
{$ENDIF}
    ProductName: array[0..63] of AnsiChar;                       ///< UTF8-encoded product identification string (e.g. "Oculus Rift DK1").
    Manufacturer: array[0..63] of AnsiChar;                      ///< UTF8-encoded HMD manufacturer identification string.
    VendorId: Int16;                                             ///< HID (USB) vendor identifier of the device.
    ProductId: Int16;                                            ///< HID (USB) product identifier of the device.
    SerialNumber: array[0..23] of AnsiChar;                      ///< HMD serial number.
    FirmwareMajor: Int16;                                        ///< HMD firmware major version.
    FirmwareMinor: Int16;                                        ///< HMD firmware minor version.
    AvailableHmdCaps: UInt32;                                    ///< Capability bits described by ovrHmdCaps which the HMD currently supports.
    DefaultHmdCaps: UInt32;                                      ///< Capability bits described by ovrHmdCaps which are default for the current Hmd.
    AvailableTrackingCaps: UInt32;                               ///< Capability bits described by ovrTrackingCaps which the system currently supports.
    DefaultTrackingCaps: UInt32;                                 ///< Capability bits described by ovrTrackingCaps which are default for the current system.
    DefaultEyeFov: array[0..Ord(ovrEye_Count)-1] of TOvrFovPort; ///< Defines the recommended FOVs for the HMD.
    MaxEyeFov: array[0..Ord(ovrEye_Count)-1] of TOvrFovPort;     ///< Defines the maximum FOVs for the HMD.
    Resolution: TOvrSizei;                                       ///< Resolution of the full HMD screen (both eyes) in pixels.
    DisplayRefreshRate: Single;                                  ///< Nominal refresh rate of the display in cycles per second at the time of HMD creation.
{$IFDEF CPU64BITS}
    pad1: array[0..3] of Byte;                                   ///< \internal struct paddding.
{$ENDIF}
  end align OVR_PTR_SIZE;
  TOvrHmdDesc = ovrHmdDesc;

  /// Used as an opaque pointer to an OVR session.
  ovrSession = ^ovrHmdStruct;
  TOvrSession = ovrSession;
  ovrHmdStruct = record
  end;



  /// Bit flags describing the current status of sensor tracking.
  ///  The values must be the same as in enum StatusBits
  ///
  /// \see ovrTrackingState
  ///
  ovrStatusBits = (
    ovrStatus_OrientationTracked    = $0001,    ///< Orientation is currently tracked (connected and in use).
    ovrStatus_PositionTracked       = $0002,    ///< Position is currently tracked (false if out of range).
    ovrStatus_EnumSize              = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrStatusBits = ovrStatusBits;

  ///  Specifies the description of a single sensor.
  ///
  /// \see ovrGetTrackerDesc
  ///
  ovrTrackerDesc = record
    FrustumHFovInRadians: Single; ///< Sensor frustum horizontal field-of-view (if present).
    FrustumVFovInRadians: Single; ///< Sensor frustum vertical field-of-view (if present).
    FrustumNearZInMeters: Single; ///< Sensor frustum near Z (if present).
    FrustumFarZInMeters: Single;  ///< Sensor frustum far Z (if present).
  end align OVR_PTR_SIZE;
  TOvrTrackerDesc = ovrTrackerDesc;

  ///  Specifies sensor flags.
  ///
  ///  /see ovrTrackerPose
  ///
  ovrTrackerFlags = (
    ovrTracker_Connected   = $0020,      ///< The sensor is present, else the sensor is absent or offline.
    ovrTracker_PoseTracked = $0004       ///< The sensor has a valid pose, else the pose is unavailable. This will only be set if ovrTracker_Connected is set.
  );
  TOvrTrackerFlags = ovrTrackerFlags;

  ///  Specifies the pose for a single sensor.
  ///
  ovrTrackerPose = record
    TrackerFlags: UInt32;      ///< ovrTrackerFlags.
    Pose: TOvrPosef;           ///< The sensor's pose. This pose includes sensor tilt (roll and pitch). For a leveled coordinate system use LeveledPose.
    LeveledPose: TOvrPosef;    ///< The sensor's leveled pose, aligned with gravity. This value includes position and yaw of the sensor, but not roll and pitch. It can be used as a reference point to render real-world objects in the correct location.
    pad0: array[0..3] of Byte; ///< \internal struct pad.
  end align 8;
  TOvrTrackerPose = ovrTrackerPose;

  /// Tracking state at a given absolute time (describes predicted HMD pose, etc.).
  /// Returned by ovr_GetTrackingState.
  ///
  /// \see ovr_GetTrackingState
  ///
  ovrTrackingState = record
    /// Predicted head pose (and derivatives) at the requested absolute time.
    HeadPose: TOvrPoseStatef;

    /// HeadPose tracking status described by ovrStatusBits.
    StatusFlags: UInt32;

    /// The most recent calculated pose for each hand when hand controller tracking is present.
    /// HandPoses[ovrHand_Left] refers to the left hand and HandPoses[ovrHand_Right] to the right hand.
    /// These values can be combined with ovrInputState for complete hand controller information.
    HandPoses: array[0..1] of TOvrPoseStatef;

    /// HandPoses status flags described by ovrStatusBits.
    /// Only ovrStatus_OrientationTracked and ovrStatus_PositionTracked are reported.
    HandStatusFlags: array[0..1] of UInt32;

    /// The pose of the origin captured during calibration.
    /// Like all other poses here, this is expressed in the space set by ovr_RecenterTrackingOrigin,
    /// and so will change every time that is called. This pose can be used to calculate
    /// where the calibrated origin lands in the new recentered space.
    /// If an application never calls ovr_RecenterTrackingOrigin, expect this value to be the identity
    /// pose and as such will point respective origin based on ovrTrackingOrigin requested when
    /// calling ovr_GetTrackingState.
    CalibratedOrigin: TOvrPosef;
  end align 8;
  TOvrTrackingState = ovrTrackingState;

  /// Rendering information for each eye. Computed by ovr_GetRenderDesc() based on the
  /// specified FOV. Note that the rendering viewport is not included
  /// here as it can be specified separately and modified per frame by
  /// passing different Viewport values in the layer structure.
  ///
  /// \see ovr_GetRenderDesc
  ///
  ovrEyeRenderDesc = record
    Eye: TOvrEyeType;                        ///< The eye index to which this instance corresponds.
    Fov: TOvrFovPort;                        ///< The field of view.
    DistortedViewport: TOvrRecti;            ///< Distortion viewport.
    PixelsPerTanAngleAtCenter: TOvrVector2f; ///< How many display pixels will fit in tan(angle) = 1.
    HmdToEyeOffset: TOvrVector3f;            ///< Translation of each eye, in meters.
  end align 4;
  TOvrEyeRenderDesc = ovrEyeRenderDesc;

  /// Projection information for ovrLayerEyeFovDepth.
  ///
  /// Use the utility function ovrTimewarpProjectionDesc_FromProjection to
  /// generate this structure from the application's projection matrix.
  ///
  /// \see ovrLayerEyeFovDepth, ovrTimewarpProjectionDesc_FromProjection
  ///
  ovrTimewarpProjectionDesc = record
    Projection22: Single; ///< Projection matrix element [2][2].
    Projection23: Single; ///< Projection matrix element [2][3].
    Projection32: Single; ///< Projection matrix element [3][2].
  end align 4;
  TOvrTimewarpProjectionDesc = ovrTimewarpProjectionDesc;

  /// Contains the data necessary to properly calculate position info for various layer types.
  /// - HmdToEyeOffset is the same value pair provided in ovrEyeRenderDesc.
  /// - HmdSpaceToWorldScaleInMeters is used to scale player motion into in-application units.
  ///   In other words, it is how big an in-application unit is in the player's physical meters.
  ///   For example, if the application uses inches as its units then HmdSpaceToWorldScaleInMeters would be 0.0254.
  ///   Note that if you are scaling the player in size, this must also scale. So if your application
  ///   units are inches, but you're shrinking the player to half their normal size, then
  ///   HmdSpaceToWorldScaleInMeters would be 0.0254*2.0.
  ///
  /// \see ovrEyeRenderDesc, ovr_SubmitFrame
  ///
  POvrViewScaleDesc = ^TOvrViewScaleDesc;
  ovrViewScaleDesc = record
    HmdToEyeOffset: array[0..Ord(ovrEye_Count)-1] of TOvrVector3f; ///< Translation of each eye.
    HmdSpaceToWorldScaleInMeters: Single;                          ///< Ratio of viewer units to meter units.
  end align 4;
  TOvrViewScaleDesc = ovrViewScaleDesc;

//-----------------------------------------------------------------------------------
// ***** Platform-independent Rendering Configuration

  /// The type of texture resource.
  ///
  /// \see ovrTextureSwapChainDesc
  ///
  ovrTextureType = (
    ovrTexture_2D,              ///< 2D textures.
    ovrTexture_2D_External,     ///< External 2D texture. Not used on PC
    ovrTexture_Cube,            ///< Cube maps. Not currently supported on PC.
    ovrTexture_Count,
    ovrTexture_EnumSize = $7fffffff  ///< \internal Force type int32_t.
  );
  TOvrTextureType = ovrTextureType;

  /// The bindings required for texture swap chain.
  ///
  /// All texture swap chains are automatically bindable as shader
  /// input resources since the Oculus runtime needs this to read them.
  ///
  /// \see ovrTextureSwapChainDesc
  ///
  ovrTextureBindFlags = (
    ovrTextureBind_None,
    ovrTextureBind_DX_RenderTarget = $0001,    ///< The application can write into the chain with pixel shader
    ovrTextureBind_DX_UnorderedAccess = $0002, ///< The application can write to the chain with compute shader
    ovrTextureBind_DX_DepthStencil = $0004,    ///< The chain buffers can be bound as depth and/or stencil buffers

    ovrTextureBind_EnumSize = $7fffffff  ///< \internal Force type int32_t.
  );
  TOvrTextureBindFlags = ovrTextureBindFlags;

  /// The format of a texture.
  ///
  /// \see ovrTextureSwapChainDesc
  ///
  ovrTextureFormat = (
    OVR_FORMAT_UNKNOWN,
    OVR_FORMAT_B5G6R5_UNORM,    ///< Not currently supported on PC. Would require a DirectX 11.1 device.
    OVR_FORMAT_B5G5R5A1_UNORM,  ///< Not currently supported on PC. Would require a DirectX 11.1 device.
    OVR_FORMAT_B4G4R4A4_UNORM,  ///< Not currently supported on PC. Would require a DirectX 11.1 device.
    OVR_FORMAT_R8G8B8A8_UNORM,
    OVR_FORMAT_R8G8B8A8_UNORM_SRGB,
    OVR_FORMAT_B8G8R8A8_UNORM,
    OVR_FORMAT_B8G8R8A8_UNORM_SRGB, ///< Not supported for OpenGL applications
    OVR_FORMAT_B8G8R8X8_UNORM,      ///< Not supported for OpenGL applications
    OVR_FORMAT_B8G8R8X8_UNORM_SRGB, ///< Not supported for OpenGL applications
    OVR_FORMAT_R16G16B16A16_FLOAT,
    OVR_FORMAT_D16_UNORM,
    OVR_FORMAT_D24_UNORM_S8_UINT,
    OVR_FORMAT_D32_FLOAT,
    OVR_FORMAT_D32_FLOAT_S8X24_UINT,

    OVR_FORMAT_ENUMSIZE = $7fffffff  ///< \internal Force type int32_t.
  );
  TOvrTextureFormat = ovrTextureFormat;

  /// Misc flags overriding particular
  ///   behaviors of a texture swap chain
  ///
  /// \see ovrTextureSwapChainDesc
  ///
  ovrTextureFlags = (
    ovrTextureMisc_None, 

    /// DX only: The underlying texture is created with a TYPELESS equivalent of the
    /// format specified in the texture desc. The SDK will still access the
    /// texture using the format specified in the texture desc, but the app can
    /// create views with different formats if this is specified.
    ovrTextureMisc_DX_Typeless = $0001,

    /// DX only: Allow generation of the mip chain on the GPU via the GenerateMips
    /// call. This flag requires that RenderTarget binding also be specified.
    ovrTextureMisc_AllowGenerateMips = $0002,

    /// Texture swap chain contains protected content, and requires
    /// HDCP connection in order to display to HMD. Also prevents
    /// mirroring or other redirection of any frame containing this contents
    ovrTextureMisc_ProtectedContent = $0004,

    ovrTextureMisc_EnumSize = $7fffffff  ///< \internal Force type int32_t.
  );
  TOvrTextureFlags = ovrTextureFlags;

  /// Description used to create a texture swap chain.
  ///
  /// \see ovr_CreateTextureSwapChainDX
  /// \see ovr_CreateTextureSwapChainGL
  ///
  ovrTextureSwapChainDesc = record
    Type_: TOvrTextureType;
    Format: TOvrTextureFormat;
    ArraySize: Int32;      ///< Only supported with ovrTexture_2D. Not supported on PC at this time.
    Width: Int32;
    Height: Int32;
    MipLevels: Int32;
    SampleCount: Int32;    ///< Current only supported on depth textures
    StaticImage: TOvrBool; ///< Not buffered in a chain. For images that don't change
    MiscFlags: UInt32;     ///< ovrTextureFlags
    BindFlags: UInt32;     ///< ovrTextureBindFlags. Not used for GL.
  end;
  TOvrTextureSwapChainDesc = ovrTextureSwapChainDesc;

  /// Description used to create a mirror texture.
  ///
  /// \see ovr_CreateMirrorTextureDX
  /// \see ovr_CreateMirrorTextureGL
  ///
  ovrMirrorTextureDesc = record
    Format: TOvrTextureFormat;
    Width: Int32;
    Height: Int32;
    MiscFlags: UInt32;      ///< ovrTextureFlags
  end;
  TOvrMirrorTextureDesc = ovrMirrorTextureDesc;

  ovrTextureSwapChain = Pointer; //^ovrTextureSwapChainData;
  TOvrTextureSwapChain = ovrTextureSwapChain;
  ovrMirrorTexture = Pointer; //^ovrMirrorTextureData;
  TOvrMirrorTexture = ovrMirrorTexture;

  //-----------------------------------------------------------------------------------

  /// Describes button input types.
  /// Button inputs are combined; that is they will be reported as pressed if they are 
  /// pressed on either one of the two devices.
  /// The ovrButton_Up/Down/Left/Right map to both XBox D-Pad and directional buttons.
  /// The ovrButton_Enter and ovrButton_Return map to Start and Back controller buttons, respectively.
  ovrButton = (
    ovrButton_A         = $00000001,
    ovrButton_B         = $00000002,
    ovrButton_RThumb    = $00000004,
    ovrButton_RShoulder = $00000008,

    ovrButton_X         = $00000100,
    ovrButton_Y         = $00000200,
    ovrButton_LThumb    = $00000400,
    ovrButton_LShoulder = $00000800,

    // Navigation through DPad.
    ovrButton_Up        = $00010000,
    ovrButton_Down      = $00020000,
    ovrButton_Left      = $00040000,
    ovrButton_Right     = $00080000,
    ovrButton_Enter     = $00100000, // Start on XBox controller.
    ovrButton_Back      = $00200000, // Back on Xbox controller.
    ovrButton_VolUp     = $00400000,  // only supported by Remote.
    ovrButton_VolDown   = $00800000,  // only supported by Remote.
    ovrButton_Home      = $01000000,

    ovrButton_EnumSize  = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrButton = ovrButton;

const
  // Bit mask of all buttons on the right Touch controller
  ovrButton_RMask     = Ord(ovrButton_A) or Ord(ovrButton_B) or Ord(ovrButton_RThumb) or Ord(ovrButton_RShoulder);

  // Bit mask of all buttons on the left Touch controller
  ovrButton_LMask     = Ord(ovrButton_X) or Ord(ovrButton_Y) or Ord(ovrButton_LThumb) or Ord(ovrButton_LShoulder);
  ovrButton_Private   = Ord(ovrButton_VolUp) or Ord(ovrButton_VolDown) or Ord(ovrButton_Home);

type
  /// Describes touch input types.
  /// These values map to capacitive touch values reported ovrInputState::Touch.
  /// Some of these values are mapped to button bits for consistency.
  ovrTouch = (
    ovrTouch_A              = Ord(ovrButton_A),
    ovrTouch_B              = Ord(ovrButton_B),
    ovrTouch_RThumb         = Ord(ovrButton_RThumb),
    ovrTouch_RIndexTrigger  = $00000010,

    ovrTouch_X              = Ord(ovrButton_X),
    ovrTouch_Y              = Ord(ovrButton_Y),
    ovrTouch_LThumb         = Ord(ovrButton_LThumb),
    ovrTouch_LIndexTrigger  = $00001000,

    // Finger pose state
    // Derived internally based on distance, proximity to sensors and filtering.
    ovrTouch_RIndexPointing = $00000020,
    ovrTouch_RThumbUp       = $00000040,

    ovrTouch_LIndexPointing = $00002000,
    ovrTouch_LThumbUp       = $00004000,

    ovrTouch_EnumSize       = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrTouch = ovrTouch;

const
  // Bit mask of all the button touches on the right controller
  ovrTouch_RButtonMask    = Ord(ovrTouch_A) or Ord(ovrTouch_B) or Ord(ovrTouch_RThumb) or Ord(ovrTouch_RIndexTrigger);
  // Bit mask of all the button touches on the left controller
  ovrTouch_LButtonMask    = Ord(ovrTouch_X) or Ord(ovrTouch_Y) or Ord(ovrTouch_LThumb) or Ord(ovrTouch_LIndexTrigger);
  // Bit mask of all right controller poses
  ovrTouch_RPoseMask      = Ord(ovrTouch_RIndexPointing) or Ord(ovrTouch_RThumbUp);
  // Bit mask of all left controller poses
  ovrTouch_LPoseMask      = Ord(ovrTouch_LIndexPointing) or Ord(ovrTouch_LThumbUp);

type
  /// Specifies which controller is connected; multiple can be connected at once.
  ovrControllerType = (
    ovrControllerType_None      = $00,
    ovrControllerType_LTouch    = $01,
    ovrControllerType_RTouch    = $02,
    ovrControllerType_Touch     = $03,
    ovrControllerType_Remote    = $04,
    ovrControllerType_XBox      = $10,

    ovrControllerType_Active    = $ff,      ///< Operate on or query whichever controller is active.

    ovrControllerType_EnumSize  = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrControllerType = ovrControllerType;

  /// Provides names for the left and right hand array indexes.
  ///
  /// \see ovrInputState, ovrTrackingState
  /// 
  ovrHandType = (
    ovrHand_Left  = 0,
    ovrHand_Right = 1,
    ovrHand_Count = 2,
    ovrHand_EnumSize = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrHandType = ovrHandType;


  /// ovrInputState describes the complete controller input state, including Oculus Touch,
  /// and XBox gamepad. If multiple inputs are connected and used at the same time,
  /// their inputs are combined.
  ovrInputState = record
    // System type when the controller state was last updated.
    TimeInSeconds: Double;

    // Values for buttons described by ovrButton.
    Buttons: UInt32;

    // Touch values for buttons and sensors as described by ovrTouch.
    Touches: UInt32;
    
    // Left and right finger trigger values (ovrHand_Left and ovrHand_Right), in the range 0.0 to 1.0f.
    IndexTrigger: array[0..1] of Single;
    
    // Left and right hand trigger values (ovrHand_Left and ovrHand_Right), in the range 0.0 to 1.0f.
    HandTrigger: array[0..1] of Single;

    // Horizontal and vertical thumbstick axis values (ovrHand_Left and ovrHand_Right), in the range -1.0f to 1.0f.
    Thumbstick: array[0..1] of TOvrVector2f;

    // The type of the controller this state is for.
    ControllerType: TOvrControllerType;
  end;
  TOvrInputState = ovrInputState;

//-----------------------------------------------------------------------------------
// ***** Initialize structures

  /// Initialization flags.
  ///
  /// \see ovrInitParams, ovr_Initialize
  ///
  ovrInitFlags = (
    /// When a debug library is requested, a slower debugging version of the library will
    /// run which can be used to help solve problems in the library and debug application code.
    ovrInit_Debug          = $00000001,

    /// When a version is requested, the LibOVR runtime respects the RequestedMinorVersion
    /// field and verifies that the RequestedMinorVersion is supported.
    ovrInit_RequestVersion = $00000004,

    // These bits are writable by user code.
    ovrinit_WritableBits   = $00ffffff,

    ovrInit_EnumSize       = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrInitFlags = ovrInitFlags;

  /// Logging levels
  ///
  /// \see ovrInitParams, ovrLogCallback
  ///
  ovrLogLevel = (
    ovrLogLevel_Debug    = 0, ///< Debug-level log event.
    ovrLogLevel_Info     = 1, ///< Info-level log event.
    ovrLogLevel_Error    = 2, ///< Error-level log event.

    ovrLogLevel_EnumSize = $7fffffff ///< \internal Force type int32_t.
  );
  TOvrLogLevel = ovrLogLevel;

  /// Signature of the logging callback function pointer type.
  ///
  /// \param[in] userData is an arbitrary value specified by the user of ovrInitParams.
  /// \param[in] level is one of the ovrLogLevel constants.
  /// \param[in] message is a UTF8-encoded null-terminated string.
  /// \see ovrInitParams, ovrLogLevel, ovr_Initialize
  ///
  ovrLogCallback = function(userData: NativeUInt; level: Int32; const message: PAnsiChar): Pointer;{$IFDEF Win32}cdecl;{$ENDIF}
  TOvrLogCallback = ovrLogCallback;

  /// Parameters for ovr_Initialize.
  ///
  /// \see ovr_Initialize
  ///
  POvrInitParams = ^TOvrInitParams;
  ovrInitParams = record
    /// Flags from ovrInitFlags to override default behavior.
    /// Use 0 for the defaults.
    Flags: UInt32;

    /// Requests a specific minimum minor version of the LibOVR runtime.
    /// Flags must include ovrInit_RequestVersion or this will be ignored
    /// and OVR_MINOR_VERSION will be used.
    RequestedMinorVersion: UInt32;

    /// User-supplied log callback function, which may be called at any time
    /// asynchronously from multiple threads until ovr_Shutdown completes.
    /// Use NULL to specify no log callback.
    LogCallback: TOvrLogCallback;

    /// User-supplied data which is passed as-is to LogCallback. Typically this 
    /// is used to store an application-specific pointer which is read in the 
    /// callback function.
    UserData: NativeUInt;

    /// Relative number of milliseconds to wait for a connection to the server
    /// before failing. Use 0 for the default timeout.
    ConnectionTimeoutMS: UInt32;

{$IFDEF CPU64BITS}
    pad0: array[0..3] of Byte; ///< \internal
{$ENDIF}
  end align 8;
  TOvrInitParams = ovrInitParams;

// -----------------------------------------------------------------------------------
// ***** API Interfaces

// Overview of the API
//
// Setup:
//  - ovr_Initialize().
//  - ovr_Create(&hmd, &graphicsId).
//  - Use hmd members and ovr_GetFovTextureSize() to determine graphics configuration
//    and ovr_GetRenderDesc() to get per-eye rendering parameters.
//  - Allocate texture swap chains with ovr_CreateTextureSwapChainDX() or
//    ovr_CreateTextureSwapChainGL(). Create any associated render target views or
//    frame buffer objects.
//
// Application Loop:
//  - Call ovr_GetPredictedDisplayTime() to get the current frame timing information.
//  - Call ovr_GetTrackingState() and ovr_CalcEyePoses() to obtain the predicted
//    rendering pose for each eye based on timing.
//  - Render the scene content into the current buffer of the texture swapchains
//    for each eye and layer you plan to update this frame. If you render into a
//    texture swap chain, you must call ovr_CommitTextureSwapChain() on it to commit
//    the changes before you reference the chain this frame (otherwise, your latest
//    changes won't be picked up).
//  - Call ovr_SubmitFrame() to render the distorted layers to and present them on the HMD.
//    If ovr_SubmitFrame returns ovrSuccess_NotVisible, there is no need to render the scene
//    for the next loop iteration. Instead, just call ovr_SubmitFrame again until it returns
//    ovrSuccess. 
//
// Shutdown:
//  - ovr_Destroy().
//  - ovr_Shutdown().


/// Initializes LibOVR
///
/// Initialize LibOVR for application usage. This includes finding and loading the LibOVRRT
/// shared library. No LibOVR API functions, other than ovr_GetLastErrorInfo and ovr_Detect, can
/// be called unless ovr_Initialize succeeds. A successful call to ovr_Initialize must be eventually
/// followed by a call to ovr_Shutdown. ovr_Initialize calls are idempotent.
/// Calling ovr_Initialize twice does not require two matching calls to ovr_Shutdown.
/// If already initialized, the return value is ovr_Success.
/// 
/// LibOVRRT shared library search order:
///      -# Current working directory (often the same as the application directory).
///      -# Module directory (usually the same as the application directory,
///         but not if the module is a separate shared library).
///      -# Application directory
///      -# Development directory (only if OVR_ENABLE_DEVELOPER_SEARCH is enabled,
///         which is off by default).
///      -# Standard OS shared library search location(s) (OS-specific).
///
/// \param params Specifies custom initialization options. May be NULL to indicate default options.
/// \return Returns an ovrResult indicating success or failure. In the case of failure, use
///         ovr_GetLastErrorInfo to get more information. Example failed results include:
///     - ovrError_Initialize: Generic initialization error.
///     - ovrError_LibLoad: Couldn't load LibOVRRT.
///     - ovrError_LibVersion: LibOVRRT version incompatibility.
///     - ovrError_ServiceConnection: Couldn't connect to the OVR Service.
///     - ovrError_ServiceVersion: OVR Service version incompatibility.
///     - ovrError_IncompatibleOS: The operating system version is incompatible.
///     - ovrError_DisplayInit: Unable to initialize the HMD display.
///     - ovrError_ServerStart:  Unable to start the server. Is it already running?
///     - ovrError_Reinitialization: Attempted to re-initialize with a different version.
///
/// <b>Example code</b>
///     \code{.cpp}
///         ovrResult result = ovr_Initialize(NULL);
///         if(OVR_FAILURE(result)) {
///             ovrErrorInfo errorInfo;
///             ovr_GetLastErrorInfo(&errorInfo);
///             DebugLog("ovr_Initialize failed: %s", errorInfo.ErrorString);
///             return false;
///         }
///         [...]
///     \endcode
///
/// \see ovr_Shutdown
///
function ovr_Initialize(params: POvrInitParams): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}


/// Shuts down LibOVR
///
/// A successful call to ovr_Initialize must be eventually matched by a call to ovr_Shutdown.
/// After calling ovr_Shutdown, no LibOVR functions can be called except ovr_GetLastErrorInfo
/// or another ovr_Initialize. ovr_Shutdown invalidates all pointers, references, and created objects
/// previously returned by LibOVR functions. The LibOVRRT shared library can be unloaded by
/// ovr_Shutdown.
///
/// \see ovr_Initialize
///
procedure ovr_Shutdown;{$IFDEF Win32}cdecl;{$ENDIF}

/// Returns information about the most recent failed return value by the
/// current thread for this library.
///
/// This function itself can never generate an error.
/// The last error is never cleared by LibOVR, but will be overwritten by new errors.
/// Do not use this call to determine if there was an error in the last API
/// call as successful API calls don't clear the last ovrErrorInfo.
/// To avoid any inconsistency, ovr_GetLastErrorInfo should be called immediately
/// after an API function that returned a failed ovrResult, with no other API
/// functions called in the interim.
///
/// \param[out] errorInfo The last ovrErrorInfo for the current thread.
///
/// \see ovrErrorInfo
///
procedure ovr_GetLastErrorInfo(out errorInfo: TOvrErrorInfo);{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns the version string representing the LibOVRRT version.
///
/// The returned string pointer is valid until the next call to ovr_Shutdown.
///
/// Note that the returned version string doesn't necessarily match the current
/// OVR_MAJOR_VERSION, etc., as the returned string refers to the LibOVRRT shared
/// library version and not the locally compiled interface version.
///
/// The format of this string is subject to change in future versions and its contents
/// should not be interpreted.
///
/// \return Returns a UTF8-encoded null-terminated version string.
///
function ovr_GetVersionString: PAnsiChar;{$IFDEF Win32}cdecl;{$ENDIF}


/// Writes a message string to the LibOVR tracing mechanism (if enabled).
///
/// This message will be passed back to the application via the ovrLogCallback if
/// it was registered.
///
/// \param[in] level One of the ovrLogLevel constants.
/// \param[in] message A UTF8-encoded null-terminated string.
/// \return returns the strlen of the message or a negative value if the message is too large.
///
/// \see ovrLogLevel, ovrLogCallback
///
function ovr_TraceMessage(level: Int32; const message: PAnsiChar): Int32;{$IFDEF Win32}cdecl;{$ENDIF}


//-------------------------------------------------------------------------------------
/// @name HMD Management
///
/// Handles the enumeration, creation, destruction, and properties of an HMD (head-mounted display).
///@{


/// Returns information about the current HMD.
///
/// ovr_Initialize must have first been called in order for this to succeed, otherwise ovrHmdDesc::Type
/// will be reported as ovrHmd_None.
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create, else NULL in which
///                case this function detects whether an HMD is present and returns its info if so.
///
/// \return Returns an ovrHmdDesc. If the hmd is NULL and ovrHmdDesc::Type is ovrHmd_None then 
///         no HMD is present.
///
function ovr_GetHmdDesc(session: TOvrSession): TOvrHmdDesc;{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns the number of sensors. 
///
/// The number of sensors may change at any time, so this function should be called before use 
/// as opposed to once on startup.
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
///
/// \return Returns unsigned int count.
///
function ovr_GetTrackerCount(session: TOvrSession): UInt32;{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns a given sensor description.
///
/// It's possible that sensor desc [0] may indicate a unconnnected or non-pose tracked sensor, but 
/// sensor desc [1] may be connected.
///
/// ovr_Initialize must have first been called in order for this to succeed, otherwise the returned
/// trackerDescArray will be zero-initialized. The data returned by this function can change at runtime.
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// 
/// \param[in] trackerDescIndex Specifies a sensor index. The valid indexes are in the range of 0 to 
///            the sensor count returned by ovr_GetTrackerCount.
///
/// \return Returns ovrTrackerDesc. An empty ovrTrackerDesc will be returned if trackerDescIndex is out of range.
///
/// \see ovrTrackerDesc, ovr_GetTrackerCount
///
function ovr_GetTrackerDesc(session: TOvrSession; trackerDescIndex: UInt32): TOvrTrackerDesc;{$IFDEF Win32}cdecl;{$ENDIF}


/// Creates a handle to a VR session.
///
/// Upon success the returned ovrSession must be eventually freed with ovr_Destroy when it is no longer needed.
/// A second call to ovr_Create will result in an error return value if the previous Hmd has not been destroyed.
///
/// \param[out] pSession Provides a pointer to an ovrSession which will be written to upon success.
/// \param[out] luid Provides a system specific graphics adapter identifier that locates which
/// graphics adapter has the HMD attached. This must match the adapter used by the application
/// or no rendering output will be possible. This is important for stability on multi-adapter systems. An
/// application that simply chooses the default adapter will not run reliably on multi-adapter systems.
/// \return Returns an ovrResult indicating success or failure. Upon failure
///         the returned pHmd will be NULL.
///
/// <b>Example code</b>
///     \code{.cpp}
///         ovrSession session;
///         ovrGraphicsLuid luid;
///         ovrResult result = ovr_Create(&session, &luid);
///         if(OVR_FAILURE(result))
///            ...
///     \endcode
///
/// \see ovr_Destroy
///
function ovr_Create(out pSession: TOvrSession; out pLuid: TOvrGraphicsLuid): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}


/// Destroys the HMD.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \see ovr_Create
///
procedure ovr_Destroy(session: TOvrSession);{$IFDEF Win32}cdecl;{$ENDIF}


/// Specifies status information for the current session.
///
/// \see ovr_GetSessionStatus
///
type
  ovrSessionStatus = record
    IsVisible: TOvrBool;      ///< True if the process has VR focus and thus is visible in the HMD.
    HmdPresent: TOvrBool;     ///< True if an HMD is present.  
    HmdMounted: TOvrBool;     ///< True if the HMD is on the user's head.
    DisplayLost: TOvrBool;    ///< True if the session is in a display-lost state. See ovr_SubmitFrame.
    ShouldQuit: TOvrBool;     ///< True if the application should initiate shutdown.    
    ShouldRecenter: TOvrBool; ///< True if UX has requested re-centering. Must call ovr_ClearShouldRecenterFlag or ovr_RecenterTrackingOrigin.
  end;
  TOvrSessionStatus = ovrSessionStatus;

/// Returns status information for the application.
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[out] sessionStatus Provides an ovrSessionStatus that is filled in.
/// 
/// \return Returns an ovrResult indicating success or failure. In the case of 
///         failure, use ovr_GetLastErrorInfo to get more information.
//          Return values include but aren't limited to:
///     - ovrSuccess: Completed successfully.
///     - ovrError_ServiceConnection: The service connection was lost and the application
//        must destroy the session.
/// 
function ovr_GetSessionStatus(session: TOvrSession; out sessionStatus: TOvrSessionStatus): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}


//-------------------------------------------------------------------------------------
/// @name Tracking
///
/// Tracking functions handle the position, orientation, and movement of the HMD in space.
///
/// All tracking interface functions are thread-safe, allowing tracking state to be sampled
/// from different threads.
///
///@{

/// Sets the tracking origin type
///
/// When the tracking origin is changed, all of the calls that either provide
/// or accept ovrPosef will use the new tracking origin provided.
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] origin Specifies an ovrTrackingOrigin to be used for all ovrPosef
///
/// \return Returns an ovrResult indicating success or failure. In the case of failure, use
///         ovr_GetLastErrorInfo to get more information.
///
/// \see ovrTrackingOrigin, ovr_GetTrackingOriginType
function ovr_SetTrackingOriginType(session: TOvrSession; origin: TOvrTrackingOrigin): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}


/// Gets the tracking origin state
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
///
/// \return Returns the ovrTrackingOrigin that was either set by default, or previous set by the application.
///
/// \see ovrTrackingOrigin, ovr_SetTrackingOriginType
function ovr_GetTrackingOriginType(session: TOvrSession): TOvrTrackingOrigin;{$IFDEF Win32}cdecl;{$ENDIF}


/// Re-centers the sensor position and orientation.
///
/// This resets the (x,y,z) positional components and the yaw orientation component.
/// The Roll and pitch orientation components are always determined by gravity and cannot
/// be redefined. All future tracking will report values relative to this new reference position.
/// If you are using ovrTrackerPoses then you will need to call ovr_GetTrackerPose after 
/// this, because the sensor position(s) will change as a result of this.
/// 
/// The headset cannot be facing vertically upward or downward but rather must be roughly
/// level otherwise this function will fail with ovrError_InvalidHeadsetOrientation.
///
/// For more info, see the notes on each ovrTrackingOrigin enumeration to understand how
/// recenter will vary slightly in its behavior based on the current ovrTrackingOrigin setting.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
///
/// \return Returns an ovrResult indicating success or failure. In the case of failure, use
///         ovr_GetLastErrorInfo to get more information. Return values include but aren't limited to:
///     - ovrSuccess: Completed successfully.
///     - ovrError_InvalidHeadsetOrientation: The headset was facing an invalid direction when
///       attempting recentering, such as facing vertically.
///
/// \see ovrTrackingOrigin, ovr_GetTrackerPose
///
function ovr_RecenterTrackingOrigin(session: TOvrSession): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}


/// Clears the ShouldRecenter status bit in ovrSessionStatus.
///
/// Clears the ShouldRecenter status bit in ovrSessionStatus, allowing further recenter 
/// requests to be detected. Since this is automatically done by ovr_RecenterTrackingOrigin,
/// this is only needs to be called when application is doing its own re-centering.
procedure ovr_ClearShouldRecenterFlag(session: TOvrSession);{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns tracking state reading based on the specified absolute system time.
///
/// Pass an absTime value of 0.0 to request the most recent sensor reading. In this case
/// both PredictedPose and SamplePose will have the same value.
///
/// This may also be used for more refined timing of front buffer rendering logic, and so on.
/// This may be called by multiple threads.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] absTime Specifies the absolute future time to predict the return
///            ovrTrackingState value. Use 0 to request the most recent tracking state.
/// \param[in] latencyMarker Specifies that this call is the point in time where
///            the "App-to-Mid-Photon" latency timer starts from. If a given ovrLayer
///            provides "SensorSampleTimestamp", that will override the value stored here.
/// \return Returns the ovrTrackingState that is predicted for the given absTime.
///
/// \see ovrTrackingState, ovr_GetEyePoses, ovr_GetTimeInSeconds
///
function ovr_GetTrackingState(session: TOvrSession; absTime: Double; latencyMarker: TOvrBool): TOvrTrackingState;{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns the ovrTrackerPose for the given sensor.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] trackerPoseIndex Index of the sensor being requested.
///
/// \return Returns the requested ovrTrackerPose. An empty ovrTrackerPose will be returned if trackerPoseIndex is out of range.
///
/// \see ovr_GetTrackerCount
///
function ovr_GetTrackerPose(session: TOvrSession; trackerPoseIndex: UInt32): TOvrTrackerPose;{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns the most recent input state for controllers, without positional tracking info.
/// Developers can tell whether the same state was returned by checking the PacketNumber.
///
/// \param[out] inputState Input state that will be filled in.
/// \param[in] ovrControllerType Specifies which controller the input will be returned for.
///            Described by ovrControllerType.
/// \return Returns ovrSuccess if the new state was successfully obtained.
///
/// \see ovrControllerType
///
function ovr_GetInputState(session: TOvrSession; controllerType: TOvrControllerType; out inputState: TOvrInputState): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns controller types connected to the system OR'ed together.
///
/// \return A bitmask of ovrControllerTypes connected to the system.
///
/// \see ovrControllerType
///
function ovr_GetConnectedControllerTypes(session: TOvrSession): UInt32;{$IFDEF Win32}cdecl;{$ENDIF}


/// Turns on vibration of the given controller.
///
/// To disable vibration, call ovr_SetControllerVibration with an amplitude of 0.
/// Vibration automatically stops after a nominal amount of time, so if you want vibration 
/// to be continuous over multiple seconds then you need to call this function periodically.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] controllerType Specifies the controller to apply the vibration to.
/// \param[in] frequency Specifies a vibration frequency in the range of 0.0 to 1.0. 
///            Currently the only valid values are 0.0, 0.5, and 1.0 and other values will
///            be clamped to one of these.
/// \param[in] amplitude Specifies a vibration amplitude in the range of 0.0 to 1.0.
///
/// \return Returns ovrSuccess upon success.
///
/// \see ovrControllerType
/// 
function ovr_SetControllerVibration(session: TOvrSession; controllerType: TOvrControllerType;
  frequency, amplitude: Single): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}

///@}


//-------------------------------------------------------------------------------------
// @name Layers
//
///@{


const
  ///  Specifies the maximum number of layers supported by ovr_SubmitFrame.
  ///
  ///  /see ovr_SubmitFrame
  ///
  ovrMaxLayerCount = 16;

type
  /// Describes layer types that can be passed to ovr_SubmitFrame.
  /// Each layer type has an associated struct, such as ovrLayerEyeFov.
  ///
  /// \see ovrLayerHeader
  ///
  ovrLayerType = (
    ovrLayerType_Disabled    = 0,        ///< Layer is disabled.
    ovrLayerType_EyeFov      = 1,        ///< Described by ovrLayerEyeFov.
    ovrLayerType_Quad        = 3,        ///< Described by ovrLayerQuad. Previously called ovrLayerType_QuadInWorld.
    /// enum 4 used to be ovrLayerType_QuadHeadLocked. Instead, use ovrLayerType_Quad with ovrLayerFlag_HeadLocked.
    ovrLayerType_EyeMatrix   = 5,        ///< Described by ovrLayerEyeMatrix.
    ovrLayerType_EnumSize    = $7fffffff ///< Force type int32_t.
  );
  TOvrLayerType = ovrLayerType;

  /// Identifies flags used by ovrLayerHeader and which are passed to ovr_SubmitFrame.
  ///
  /// \see ovrLayerHeader
  ///
  ovrLayerFlags = (
    /// ovrLayerFlag_HighQuality enables 4x anisotropic sampling during the composition of the layer.
    /// The benefits are mostly visible at the periphery for high-frequency & high-contrast visuals.
    /// For best results consider combining this flag with an ovrTextureSwapChain that has mipmaps and
    /// instead of using arbitrary sized textures, prefer texture sizes that are powers-of-two.
    /// Actual rendered viewport and doesn't necessarily have to fill the whole texture.
    ovrLayerFlag_HighQuality               = $01,

    /// ovrLayerFlag_TextureOriginAtBottomLeft: the opposite is TopLeft.
    /// Generally this is false for D3D, true for OpenGL.
    ovrLayerFlag_TextureOriginAtBottomLeft = $02,

    /// Mark this surface as "headlocked", which means it is specified
    /// relative to the HMD and moves with it, rather than being specified
    /// relative to sensor/torso space and remaining still while the head moves.
    /// What used to be ovrLayerType_QuadHeadLocked is now ovrLayerType_Quad plus this flag.
    /// However the flag can be applied to any layer type to achieve a similar effect.
    ovrLayerFlag_HeadLocked                = $04
  );
  TOvrLayerFlags = ovrLayerFlags;

  /// Defines properties shared by all ovrLayer structs, such as ovrLayerEyeFov.
  ///
  /// ovrLayerHeader is used as a base member in these larger structs.
  /// This struct cannot be used by itself except for the case that Type is ovrLayerType_Disabled.
  ///
  /// \see ovrLayerType, ovrLayerFlags
  ///
  ovrLayerHeader = record
    Type_: TOvrLayerType; ///< Described by ovrLayerType.
    Flags: UInt32;        ///< Described by ovrLayerFlags.
  end align OVR_PTR_SIZE;
  POvrLayerHeader = ^TOvrLayerHeader;
  TOvrLayerHeader = ovrLayerHeader;
  TOvrLayerHeaderArray = array[0..0] of TOvrLayerHeader;
  POvrLayerHeaderArray = ^TOvrLayerHeaderArray;

  /// Describes a layer that specifies a monoscopic or stereoscopic view.
  /// This is the kind of layer that's typically used as layer 0 to ovr_SubmitFrame,
  /// as it is the kind of layer used to render a 3D stereoscopic view.
  ///
  /// Three options exist with respect to mono/stereo texture usage:
  ///    - ColorTexture[0] and ColorTexture[1] contain the left and right stereo renderings, respectively.
  ///      Viewport[0] and Viewport[1] refer to ColorTexture[0] and ColorTexture[1], respectively.
  ///    - ColorTexture[0] contains both the left and right renderings, ColorTexture[1] is NULL,
  ///      and Viewport[0] and Viewport[1] refer to sub-rects with ColorTexture[0].
  ///    - ColorTexture[0] contains a single monoscopic rendering, and Viewport[0] and
  ///      Viewport[1] both refer to that rendering.
  ///
  /// \see ovrTextureSwapChain, ovr_SubmitFrame
  ///
  ovrLayerEyeFov = record
    /// Header.Type must be ovrLayerType_EyeFov.
    Header: TOvrLayerHeader;

    /// ovrTextureSwapChains for the left and right eye respectively.
    /// The second one of which can be NULL for cases described above.
    ColorTexture: array[0..Ord(ovrEye_Count)-1] of TOvrTextureSwapChain;

    /// Specifies the ColorTexture sub-rect UV coordinates.
    /// Both Viewport[0] and Viewport[1] must be valid.
    Viewport: array[0..Ord(ovrEye_Count)-1] of TOvrRecti;

    /// The viewport field of view.
    Fov: array[0..Ord(ovrEye_Count)-1] of TOvrFovPort;

    /// Specifies the position and orientation of each eye view, with the position specified in meters.
    /// RenderPose will typically be the value returned from ovr_CalcEyePoses,
    /// but can be different in special cases if a different head pose is used for rendering.
    RenderPose: array[0..Ord(ovrEye_Count)-1] of TOvrPosef;

    /// Specifies the timestamp when the source ovrPosef (used in calculating RenderPose)
    /// was sampled from the SDK. Typically retrieved by calling ovr_GetTimeInSeconds
    /// around the instant the application calls ovr_GetTrackingState
    /// The main purpose for this is to accurately track app tracking latency.
    SensorSampleTime: Double;

  end align OVR_PTR_SIZE;
  TOvrLayerEyeFov = ovrLayerEyeFov;

  /// Describes a layer that specifies a monoscopic or stereoscopic view.
  /// This uses a direct 3x4 matrix to map from view space to the UV coordinates.
  /// It is essentially the same thing as ovrLayerEyeFov but using a much
  /// lower level. This is mainly to provide compatibility with specific apps.
  /// Unless the application really requires this flexibility, it is usually better
  /// to use ovrLayerEyeFov.
  ///
  /// Three options exist with respect to mono/stereo texture usage:
  ///    - ColorTexture[0] and ColorTexture[1] contain the left and right stereo renderings, respectively.
  ///      Viewport[0] and Viewport[1] refer to ColorTexture[0] and ColorTexture[1], respectively.
  ///    - ColorTexture[0] contains both the left and right renderings, ColorTexture[1] is NULL,
  ///      and Viewport[0] and Viewport[1] refer to sub-rects with ColorTexture[0].
  ///    - ColorTexture[0] contains a single monoscopic rendering, and Viewport[0] and
  ///      Viewport[1] both refer to that rendering.
  ///
  /// \see ovrTextureSwapChain, ovr_SubmitFrame
  ///
  ovrLayerEyeMatrix = record
    /// Header.Type must be ovrLayerType_EyeMatrix.
    Header: TOvrLayerHeader;

    /// ovrTextureSwapChains for the left and right eye respectively.
    /// The second one of which can be NULL for cases described above.
    ColorTexture: array[0..Ord(ovrEye_Count)-1] of TOvrTextureSwapChain;

    /// Specifies the ColorTexture sub-rect UV coordinates.
    /// Both Viewport[0] and Viewport[1] must be valid.
    Viewport: array[0..Ord(ovrEye_Count)-1] of TOvrRecti;

    /// Specifies the position and orientation of each eye view, with the position specified in meters.
    /// RenderPose will typically be the value returned from ovr_CalcEyePoses,
    /// but can be different in special cases if a different head pose is used for rendering.
    RenderPose: array[0..Ord(ovrEye_Count)-1] of TOvrPosef;

    /// Specifies the mapping from a view-space vector
    /// to a UV coordinate on the textures given above.
    /// P = (x,y,z,1)*Matrix
    /// TexU  = P.x/P.z
    /// TexV  = P.y/P.z
    Matrix: array[0..Ord(ovrEye_Count)-1] of TOvrMatrix4f;

    /// Specifies the timestamp when the source ovrPosef (used in calculating RenderPose)
    /// was sampled from the SDK. Typically retrieved by calling ovr_GetTimeInSeconds
    /// around the instant the application calls ovr_GetTrackingState
    /// The main purpose for this is to accurately track app tracking latency.
    SensorSampleTime: Double;

  end align OVR_PTR_SIZE;
  TOvrLayerEyeMatrix = ovrLayerEyeMatrix;




  /// Describes a layer of Quad type, which is a single quad in world or viewer space.
  /// It is used for ovrLayerType_Quad. This type of layer represents a single
  /// object placed in the world and not a stereo view of the world itself.
  ///
  /// A typical use of ovrLayerType_Quad is to draw a television screen in a room
  /// that for some reason is more convenient to draw as a layer than as part of the main
  /// view in layer 0. For example, it could implement a 3D popup GUI that is drawn at a
  /// higher resolution than layer 0 to improve fidelity of the GUI.
  ///
  /// Quad layers are visible from both sides; they are not back-face culled.
  ///
  /// \see ovrTextureSwapChain, ovr_SubmitFrame
  ///
  ovrLayerQuad = record
    /// Header.Type must be ovrLayerType_Quad.
    Header: TOvrLayerHeader;

    /// Contains a single image, never with any stereo view.
    ColorTexture: TOvrTextureSwapChain;

    /// Specifies the ColorTexture sub-rect UV coordinates.
    Viewport: TOvrRecti;

    /// Specifies the orientation and position of the center point of a Quad layer type.
    /// The supplied direction is the vector perpendicular to the quad.
    /// The position is in real-world meters (not the application's virtual world,
    /// the physical world the user is in) and is relative to the "zero" position
    /// set by ovr_RecenterTrackingOrigin unless the ovrLayerFlag_HeadLocked flag is used.
    QuadPoseCenter: TOvrPosef;

    /// Width and height (respectively) of the quad in meters.
    QuadSize: TOvrVector2f;

  end align OVR_PTR_SIZE;
  TOvrLayerQuad = ovrLayerQuad;

  /// Union that combines ovrLayer types in a way that allows them
  /// to be used in a polymorphic way.
  ovrLayer_Union = record
    Header: TOvrLayerHeader;
    EyeFov: TOvrLayerEyeFov;
    Quad: TOvrLayerQuad;
  end;
  TOvrLayer_Union = ovrLayer_Union;

//@}



/// @name SDK Distortion Rendering
///
/// All of rendering functions including the configure and frame functions
/// are not thread safe. It is OK to use ConfigureRendering on one thread and handle
/// frames on another thread, but explicit synchronization must be done since
/// functions that depend on configured state are not reentrant.
///
/// These functions support rendering of distortion by the SDK.
///
//@{

/// TextureSwapChain creation is rendering API-specific.
/// ovr_CreateTextureSwapChainDX and ovr_CreateTextureSwapChainGL can be found in the
/// rendering API-specific headers, such as OVR_CAPI_D3D.h and OVR_CAPI_GL.h

/// Gets the number of buffers in an ovrTextureSwapChain.
///
/// \param[in]  session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in]  chain Specifies the ovrTextureSwapChain for which the length should be retrieved.
/// \param[out] out_Length Returns the number of buffers in the specified chain.
///
/// \return Returns an ovrResult for which OVR_SUCCESS(result) is false upon error. 
///
/// \see ovr_CreateTextureSwapChainDX, ovr_CreateTextureSwapChainGL
///
function ovr_GetTextureSwapChainLength(session: TOvrSession; chain: TOvrTextureSwapChain; out out_Length: Int32): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}

/// Gets the current index in an ovrTextureSwapChain.
///
/// \param[in]  session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in]  chain Specifies the ovrTextureSwapChain for which the index should be retrieved.
/// \param[out] out_Index Returns the current (free) index in specified chain.
///
/// \return Returns an ovrResult for which OVR_SUCCESS(result) is false upon error. 
///
/// \see ovr_CreateTextureSwapChainDX, ovr_CreateTextureSwapChainGL
///
function ovr_GetTextureSwapChainCurrentIndex(session: TOvrSession; chain: TOvrTextureSwapChain; out out_Index: Int32): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}

/// Gets the description of the buffers in an ovrTextureSwapChain
///
/// \param[in]  session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in]  chain Specifies the ovrTextureSwapChain for which the description should be retrieved.
/// \param[out] out_Desc Returns the description of the specified chain.
///
/// \return Returns an ovrResult for which OVR_SUCCESS(result) is false upon error. 
///
/// \see ovr_CreateTextureSwapChainDX, ovr_CreateTextureSwapChainGL
///
function ovr_GetTextureSwapChainDesc(session: TOvrSession; chain: TOvrTextureSwapChain; out out_Desc: TOvrTextureSwapChainDesc): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}

/// Commits any pending changes to an ovrTextureSwapChain, and advances its current index
///
/// \param[in]  session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in]  chain Specifies the ovrTextureSwapChain to commit.
///
/// \note When Commit is called, the texture at the current index is considered ready for use by the
/// runtime, and further writes to it should be avoided. The swap chain's current index is advanced,
/// providing there's room in the chain. The next time the SDK dereferences this texture swap chain,
/// it will synchronize with the app's graphics context and pick up the submitted index, opening up
/// room in the swap chain for further commits.
///
/// \return Returns an ovrResult for which OVR_SUCCESS(result) is false upon error. 
///         Failures include but aren't limited to:
///     - ovrError_TextureSwapChainFull: ovr_CommitTextureSwapChain was called too many times on a texture swapchain without calling submit to use the chain.
///
/// \see ovr_CreateTextureSwapChainDX, ovr_CreateTextureSwapChainGL
///
function ovr_CommitTextureSwapChain(session: TOvrSession; chain: TOvrTextureSwapChain): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}

/// Destroys an ovrTextureSwapChain and frees all the resources associated with it.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] chain Specifies the ovrTextureSwapChain to destroy. If it is NULL then this function has no effect.
///
/// \see ovr_CreateTextureSwapChainDX, ovr_CreateTextureSwapChainGL
///
procedure ovr_DestroyTextureSwapChain(session: TOvrSession; chain: TOvrTextureSwapChain);{$IFDEF Win32}cdecl;{$ENDIF}


/// MirrorTexture creation is rendering API-specific.
/// ovr_CreateMirrorTextureDX and ovr_CreateMirrorTextureGL can be found in the
/// rendering API-specific headers, such as OVR_CAPI_D3D.h and OVR_CAPI_GL.h

/// Destroys a mirror texture previously created by one of the mirror texture creation functions.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] mirrorTexture Specifies the ovrTexture to destroy. If it is NULL then this function has no effect.
///
/// \see ovr_CreateMirrorTextureDX, ovr_CreateMirrorTextureGL
///
procedure ovr_DestroyMirrorTexture(session: TOvrSession; mirrorTexture: TOvrMirrorTexture);{$IFDEF Win32}cdecl;{$ENDIF}


/// Calculates the recommended viewport size for rendering a given eye within the HMD
/// with a given FOV cone.
///
/// Higher FOV will generally require larger textures to maintain quality.
/// Apps packing multiple eye views together on the same texture should ensure there are
/// at least 8 pixels of padding between them to prevent texture filtering and chromatic
/// aberration causing images to leak between the two eye views.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] eye Specifies which eye (left or right) to calculate for.
/// \param[in] fov Specifies the ovrFovPort to use.
///
/// \param[in] pixelsPerDisplayPixel Specifies the ratio of the number of render target pixels
///            to display pixels at the center of distortion. 1.0 is the default value. Lower
///            values can improve performance, higher values give improved quality.
///
/// <b>Example code</b>
///     \code{.cpp}
///         ovrHmdDesc hmdDesc = ovr_GetHmdDesc(session);
///         ovrSizei eyeSizeLeft  = ovr_GetFovTextureSize(session, ovrEye_Left,  hmdDesc.DefaultEyeFov[ovrEye_Left],  1.0f);
///         ovrSizei eyeSizeRight = ovr_GetFovTextureSize(session, ovrEye_Right, hmdDesc.DefaultEyeFov[ovrEye_Right], 1.0f);
///     \endcode
///
/// \return Returns the texture width and height size.
///
function ovr_GetFovTextureSize(session: TOvrSession; eye: TOvrEyeType; fov: TOvrFovPort;
  pixelsPerDisplayPixel: Single): TOvrSizei;{$IFDEF Win32}cdecl;{$ENDIF}

/// Computes the distortion viewport, view adjust, and other rendering parameters for
/// the specified eye.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] eyeType Specifies which eye (left or right) for which to perform calculations.
/// \param[in] fov Specifies the ovrFovPort to use.
/// \return Returns the computed ovrEyeRenderDesc for the given eyeType and field of view.
///
/// \see ovrEyeRenderDesc
///
function ovr_GetRenderDesc(session: TOvrSession; eyeType: TOvrEyeType;
  fov: TOvrFovPort): TOvrEyeRenderDesc;{$IFDEF Win32}cdecl;{$ENDIF}

/// Submits layers for distortion and display.
///
/// ovr_SubmitFrame triggers distortion and processing which might happen asynchronously.
/// The function will return when there is room in the submission queue and surfaces
/// are available. Distortion might or might not have completed.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
///
/// \param[in] frameIndex Specifies the targeted application frame index, or 0 to refer to one frame
///        after the last time ovr_SubmitFrame was called.
///
/// \param[in] viewScaleDesc Provides additional information needed only if layerPtrList contains
///        an ovrLayerType_Quad. If NULL, a default version is used based on the current configuration and a 1.0 world scale.
///
/// \param[in] layerPtrList Specifies a list of ovrLayer pointers, which can include NULL entries to
///        indicate that any previously shown layer at that index is to not be displayed.
///        Each layer header must be a part of a layer structure such as ovrLayerEyeFov or ovrLayerQuad,
///        with Header.Type identifying its type. A NULL layerPtrList entry in the array indicates the
//         absence of the given layer.
///
/// \param[in] layerCount Indicates the number of valid elements in layerPtrList. The maximum
///        supported layerCount is not currently specified, but may be specified in a future version.
///
/// - Layers are drawn in the order they are specified in the array, regardless of the layer type.
///
/// - Layers are not remembered between successive calls to ovr_SubmitFrame. A layer must be
///   specified in every call to ovr_SubmitFrame or it won't be displayed.
///
/// - If a layerPtrList entry that was specified in a previous call to ovr_SubmitFrame is
///   passed as NULL or is of type ovrLayerType_Disabled, that layer is no longer displayed.
///
/// - A layerPtrList entry can be of any layer type and multiple entries of the same layer type
///   are allowed. No layerPtrList entry may be duplicated (i.e. the same pointer as an earlier entry).
///
/// <b>Example code</b>
///     \code{.cpp}
///         ovrLayerEyeFov  layer0;
///         ovrLayerQuad    layer1;
///           ...
///         ovrLayerHeader* layers[2] = { &layer0.Header, &layer1.Header };
///         ovrResult result = ovr_SubmitFrame(hmd, frameIndex, nullptr, layers, 2);
///     \endcode
///
/// \return Returns an ovrResult for which OVR_SUCCESS(result) is false upon error and true
///         upon success. Return values include but aren't limited to:
///     - ovrSuccess: rendering completed successfully.
///     - ovrSuccess_NotVisible: rendering completed successfully but was not displayed on the HMD,
///       usually because another application currently has ownership of the HMD. Applications receiving
///       this result should stop rendering new content, but continue to call ovr_SubmitFrame periodically
///       until it returns a value other than ovrSuccess_NotVisible.
///     - ovrError_DisplayLost: The session has become invalid (such as due to a device removal)
///       and the shared resources need to be released (ovr_DestroyTextureSwapChain), the session needs to
///       destroyed (ovr_Destroy) and recreated (ovr_Create), and new resources need to be created
///       (ovr_CreateTextureSwapChainXXX). The application's existing private graphics resources do not
///       need to be recreated unless the new ovr_Create call returns a different GraphicsLuid.
///     - ovrError_TextureSwapChainInvalid: The ovrTextureSwapChain is in an incomplete or inconsistent state. 
///       Ensure ovr_CommitTextureSwapChain was called at least once first.
///
/// \see ovr_GetPredictedDisplayTime, ovrViewScaleDesc, ovrLayerHeader
///
function ovr_SubmitFrame(session: TOvrSession; frameIndex: Int64;
  const viewScaleDesc: POvrViewScaleDesc;
  const layerPtrList: POvrLayerHeaderArray; layerCount: UInt32): TOvrResult;{$IFDEF Win32}cdecl;{$ENDIF}
///@}



//-------------------------------------------------------------------------------------
/// @name Frame Timing
///
//@{


/// Gets the time of the specified frame midpoint.
///
/// Predicts the time at which the given frame will be displayed. The predicted time 
/// is the middle of the time period during which the corresponding eye images will 
/// be displayed. 
///
/// The application should increment frameIndex for each successively targeted frame,
/// and pass that index to any relevent OVR functions that need to apply to the frame
/// identified by that index.
///
/// This function is thread-safe and allows for multiple application threads to target
/// their processing to the same displayed frame.
/// 
/// In the even that prediction fails due to various reasons (e.g. the display being off
/// or app has yet to present any frames), the return value will be current CPU time.
/// 
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] frameIndex Identifies the frame the caller wishes to target.
///            A value of zero returns the next frame index.
/// \return Returns the absolute frame midpoint time for the given frameIndex.
/// \see ovr_GetTimeInSeconds
///
function ovr_GetPredictedDisplayTime(session: TOvrSession; frameIndex: Int64): Double;{$IFDEF Win32}cdecl;{$ENDIF}


/// Returns global, absolute high-resolution time in seconds.
///
/// The time frame of reference for this function is not specified and should not be
/// depended upon.
///
/// \return Returns seconds as a floating point value.
/// \see ovrPoseStatef, ovrFrameTiming
///
function ovr_GetTimeInSeconds: Double;{$IFDEF Win32}cdecl;{$ENDIF}


type
  /// Performance HUD enables the HMD user to see information critical to
  /// the real-time operation of the VR application such as latency timing,
  /// and CPU & GPU performance metrics
  ///
  ///     App can toggle performance HUD modes as such:
  ///     \code{.cpp}
  ///         ovrPerfHudMode PerfHudMode = ovrPerfHud_LatencyTiming;
  ///         ovr_SetInt(Hmd, OVR_PERF_HUD_MODE, (int)PerfHudMode);
  ///     \endcode
  ///
  ovrPerfHudMode = (
    ovrPerfHud_Off                = 0,  ///< Turns off the performance HUD
    ovrPerfHud_PerfSummary        = 1,  ///< Shows performance summary and headroom
    ovrPerfHud_LatencyTiming      = 2,  ///< Shows latency related timing info
    ovrPerfHud_AppRenderTiming    = 3,  ///< Shows render timing info for application
    ovrPerfHud_CompRenderTiming   = 4,  ///< Shows render timing info for OVR compositor
    ovrPerfHud_VersionInfo        = 5,  ///< Shows SDK & HMD version Info
    ovrPerfHud_Count              = 6,  ///< \internal Count of enumerated elements.
    ovrPerfHud_EnumSize = $7fffffff     ///< \internal Force type int32_t.
  );
  TOvrPerfHudMode = ovrPerfHudMode;

  /// Layer HUD enables the HMD user to see information about a layer
  ///
  ///     App can toggle layer HUD modes as such:
  ///     \code{.cpp}
  ///         ovrLayerHudMode LayerHudMode = ovrLayerHud_Info;
  ///         ovr_SetInt(Hmd, OVR_LAYER_HUD_MODE, (int)LayerHudMode);
  ///     \endcode
  ///
  ovrLayerHudMode = (
    ovrLayerHud_Off      = 0, ///< Turns off the layer HUD
    ovrLayerHud_Info     = 1, ///< Shows info about a specific layer
    ovrLayerHud_EnumSize = $7fffffff
  );
  TOvrLayerHudMode = ovrLayerHudMode;

///@}

  /// Debug HUD is provided to help developers gauge and debug the fidelity of their app's
  /// stereo rendering characteristics. Using the provided quad and crosshair guides, 
  /// the developer can verify various aspects such as VR tracking units (e.g. meters),
  /// stereo camera-parallax properties (e.g. making sure objects at infinity are rendered
  /// with the proper separation), measuring VR geometry sizes and distances and more.
  ///
  ///     App can toggle the debug HUD modes as such:
  ///     \code{.cpp}
  ///         ovrDebugHudStereoMode DebugHudMode = ovrDebugHudStereo_QuadWithCrosshair;
  ///         ovr_SetInt(Hmd, OVR_DEBUG_HUD_STEREO_MODE, (int)DebugHudMode);
  ///     \endcode
  ///
  /// The app can modify the visual properties of the stereo guide (i.e. quad, crosshair)
  /// using the ovr_SetFloatArray function. For a list of tweakable properties,
  /// see the OVR_DEBUG_HUD_STEREO_GUIDE_* keys in the OVR_CAPI_Keys.h header file.
  ovrDebugHudStereoMode = (
    ovrDebugHudStereo_Off                 = 0,  ///< Turns off the Stereo Debug HUD
    ovrDebugHudStereo_Quad                = 1,  ///< Renders Quad in world for Stereo Debugging
    ovrDebugHudStereo_QuadWithCrosshair   = 2,  ///< Renders Quad+crosshair in world for Stereo Debugging
    ovrDebugHudStereo_CrosshairAtInfinity = 3,  ///< Renders screen-space crosshair at infinity for Stereo Debugging
    ovrDebugHudStereo_Count,                    ///< \internal Count of enumerated elements

    ovrDebugHudStereo_EnumSize = $7fffffff      ///< \internal Force type int32_t
  );
  TOvrDebugHudStereoMode = ovrDebugHudStereoMode;




// -----------------------------------------------------------------------------------
/// @name Property Access
///
/// These functions read and write OVR properties. Supported properties
/// are defined in OVR_CAPI_Keys.h
///
//@{

/// Reads a boolean property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid for only the call.
/// \param[in] defaultVal specifes the value to return if the property couldn't be read.
/// \return Returns the property interpreted as a boolean value. Returns defaultVal if
///         the property doesn't exist.
function ovr_GetBool(session: TOvrSession; const propertyName: PAnsiChar; defaultVal: TOvrBool): TOvrBool;{$IFDEF Win32}cdecl;{$ENDIF}

/// Writes or creates a boolean property.
/// If the property wasn't previously a boolean property, it is changed to a boolean property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] value The value to write.
/// \return Returns true if successful, otherwise false. A false result should only occur if the property
///         name is empty or if the property is read-only.
function ovr_SetBool(session: TOvrSession; const propertyName: PAnsiChar; value: TOvrBool): TOvrBool;{$IFDEF Win32}cdecl;{$ENDIF}


/// Reads an integer property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] defaultVal Specifes the value to return if the property couldn't be read.
/// \return Returns the property interpreted as an integer value. Returns defaultVal if
///         the property doesn't exist.
function ovr_GetInt(session: TOvrSession; const propertyName: PAnsiChar; defaultVal: Int32): Int32;{$IFDEF Win32}cdecl;{$ENDIF}

/// Writes or creates an integer property.
///
/// If the property wasn't previously a boolean property, it is changed to an integer property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] value The value to write.
/// \return Returns true if successful, otherwise false. A false result should only occur if the property
///         name is empty or if the property is read-only.
function ovr_SetInt(session: TOvrSession; const propertyName: PAnsiChar; value: Int32): TOvrBool;{$IFDEF Win32}cdecl;{$ENDIF}


/// Reads a float property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] defaultVal specifes the value to return if the property couldn't be read.
/// \return Returns the property interpreted as an float value. Returns defaultVal if
///         the property doesn't exist.
function ovr_GetFloat(session: TOvrSession; const propertyName: PAnsiChar; defaultVal: Single): Single;{$IFDEF Win32}cdecl;{$ENDIF}

/// Writes or creates a float property.
/// If the property wasn't previously a float property, it's changed to a float property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] value The value to write.
/// \return Returns true if successful, otherwise false. A false result should only occur if the property
///         name is empty or if the property is read-only.
function ovr_SetFloat(session: TOvrSession; const propertyName: PAnsiChar; value: Single): TOvrBool;{$IFDEF Win32}cdecl;{$ENDIF}


/// Reads a float array property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] values An array of float to write to.
/// \param[in] valuesCapacity Specifies the maximum number of elements to write to the values array.
/// \return Returns the number of elements read, or 0 if property doesn't exist or is empty.
function ovr_GetFloatArray(session: TOvrSession; const propertyName: PAnsiChar;
  values: PSingle; valuesCapacity: UInt32): UInt32;{$IFDEF Win32}cdecl;{$ENDIF}

/// Writes or creates a float array property.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] values An array of float to write from.
/// \param[in] valuesSize Specifies the number of elements to write.
/// \return Returns true if successful, otherwise false. A false result should only occur if the property
///         name is empty or if the property is read-only.
function ovr_SetFloatArray(session: TOvrSession; const propertyName: PAnsiChar;
  const values: PSingle; valuesSize: UInt32): TOvrBool;{$IFDEF Win32}cdecl;{$ENDIF}


/// Reads a string property.
/// Strings are UTF8-encoded and null-terminated.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] defaultVal Specifes the value to return if the property couldn't be read.
/// \return Returns the string property if it exists. Otherwise returns defaultVal, which can be specified as NULL.
///         The return memory is guaranteed to be valid until next call to ovr_GetString or
///         until the HMD is destroyed, whichever occurs first.
function ovr_GetString(session: TOvrSession; const propertyName: PAnsiChar;
  const defaultVal: PAnsiChar): PAnsiChar;{$IFDEF Win32}cdecl;{$ENDIF}

/// Writes or creates a string property.
/// Strings are UTF8-encoded and null-terminated.
///
/// \param[in] session Specifies an ovrSession previously returned by ovr_Create.
/// \param[in] propertyName The name of the property, which needs to be valid only for the call.
/// \param[in] value The string property, which only needs to be valid for the duration of the call.
/// \return Returns true if successful, otherwise false. A false result should only occur if the property
///         name is empty or if the property is read-only.
function ovr_SetString(session: TOvrSession; const propertyName: PAnsiChar;
  const value: PAnsiChar): TOvrBool;{$IFDEF Win32}cdecl;{$ENDIF}

///@}


/// @cond DoxygenIgnore
//-----------------------------------------------------------------------------
// ***** Compiler packing validation
//
// These checks ensure that the compiler settings being used will be compatible
// with with pre-built dynamic library provided with the runtime.

{$IF SizeOf(ovrBool) <> 1}{$MESSAGE Error 'ovrBool size mismatch'}{$IFEND}
{$IF SizeOf(ovrVector2i) <> 4 * 2}{$MESSAGE Error 'ovrVector2i size mismatch'}{$IFEND}
{$IF SizeOf(ovrSizei) <> 4 * 2}{$MESSAGE Error 'ovrSizei size mismatch'}{$IFEND}
{$IF SizeOf(ovrRecti) <> SizeOf(ovrVector2i) + SizeOf(ovrSizei)}{$MESSAGE Error 'ovrRecti size mismatch'}{$IFEND}
{$IF SizeOf(ovrQuatf) <> 4 * 4}{$MESSAGE Error 'ovrQuatf size mismatch'}{$IFEND}
{$IF SizeOf(ovrVector2f) <> 4 * 2}{$MESSAGE Error 'ovrVector2f size mismatch'}{$IFEND}
{$IF SizeOf(ovrVector3f) <> 4 * 3}{$MESSAGE Error 'ovrVector3f size mismatch'}{$IFEND}
{$IF SizeOf(ovrMatrix4f) <> 4 * 16}{$MESSAGE Error 'ovrMatrix4f size mismatch'}{$IFEND}

{$IF SizeOf(ovrPosef) <> (7 * 4)}{$MESSAGE Error 'ovrPosef size mismatch'}{$IFEND}
{$IF SizeOf(ovrPoseStatef) <> (22 * 4)}{$MESSAGE Error 'ovrPoseStatef size mismatch'}{$IFEND}
{$IF SizeOf(ovrFovPort) <> (4 * 4)}{$MESSAGE Error 'ovrFovPort size mismatch'}{$IFEND}

{$IF SizeOf(ovrHmdCaps) <> 4}{$MESSAGE Error 'ovrHmdCaps size mismatch'}{$IFEND}
{$IF SizeOf(ovrTrackingCaps) <> 4}{$MESSAGE Error 'ovrTrackingCaps size mismatch'}{$IFEND}
{$IF SizeOf(ovrEyeType) <> 4}{$MESSAGE Error 'ovrEyeType size mismatch'}{$IFEND}
{$IF SizeOf(ovrHmdType) <> 4}{$MESSAGE Error 'ovrHmdType size mismatch'}{$IFEND}

{$IF SizeOf(ovrTrackerDesc) <> 4 + 4 + 4 + 4}{$MESSAGE Error 'ovrTrackerDesc size mismatch'}{$IFEND}
{$IF SizeOf(ovrTrackerPose) <> 4 + 4 + SizeOf(ovrPosef) + SizeOf(ovrPosef)}{$MESSAGE Error 'ovrTrackerPose size mismatch'}{$IFEND}
{$IF SizeOf(ovrTrackingState) <> SizeOf(ovrPoseStatef) + 4 + 4 + (SizeOf(ovrPoseStatef) * 2) + (SizeOf(Int32) * 2) + SizeOf(ovrPosef) + 4}{$MESSAGE Error 'ovrTrackingState size mismatch'}{$IFEND}


{$IF SizeOf(ovrStatusBits) <> 4}{$MESSAGE Error 'ovrStatusBits size mismatch'}{$IFEND}

{$IF SizeOf(ovrSessionStatus) <> 6}{$MESSAGE Error 'ovrSessionStatus size mismatch'}{$IFEND}

{$IF SizeOf(ovrEyeRenderDesc) <> SizeOf(ovrEyeType) + SizeOf(ovrFovPort) + SizeOf(ovrRecti) +
                                                  SizeOf(ovrVector2f) + SizeOf(ovrVector3f)}
                      {$MESSAGE Error 'ovrEyeRenderDesc size mismatch'}{$IFEND}
{$IF SizeOf(ovrTimewarpProjectionDesc) <> 4 * 3}{$MESSAGE Error 'ovrTimewarpProjectionDesc size mismatch'}{$IFEND}

{$IF SizeOf(ovrInitFlags) <> 4}{$MESSAGE Error 'ovrInitFlags size mismatch'}{$IFEND}
{$IF SizeOf(ovrLogLevel) <> 4}{$MESSAGE Error 'ovrLogLevel size mismatch'}{$IFEND}

{$IF SizeOf(ovrInitParams) <> 4 + 4 + SizeOf(ovrLogCallback) + SizeOf(NativeUInt) + 4 + 4}
                      {$MESSAGE Error 'ovrInitParams size mismatch'}{$IFEND}

{$IFDEF CPU64BITS}
  {$IF SizeOf(ovrHmdDesc) <> 
    + SizeOf(ovrHmdType)                // Type
    + 4                                 // pad0
    + 64                                // ProductName 
    + 64                                // Manufacturer
    + 2                                 // VendorId
    + 2                                 // ProductId
    + 24                                // SerialNumber
    + 2                                 // FirmwareMajor
    + 2                                 // FirmwareMinor
    + 4 * 4                             // AvailableHmdCaps - DefaultTrackingCaps
    + SizeOf(ovrFovPort) * 2            // DefaultEyeFov
    + SizeOf(ovrFovPort) * 2            // MaxEyeFov
    + SizeOf(ovrSizei)                  // Resolution
    + 4                                 // DisplayRefreshRate
    + 4                                 // pad1
    }{$MESSAGE Error 'ovrHmdDesc size mismatch'}{$IFEND}
{$ELSE}
  {$IF SizeOf(ovrHmdDesc) <> 
    + SizeOf(ovrHmdType)                // Type
    + 64                                // ProductName 
    + 64                                // Manufacturer
    + 2                                 // VendorId
    + 2                                 // ProductId
    + 24                                // SerialNumber
    + 2                                 // FirmwareMajor
    + 2                                 // FirmwareMinor
    + 4 * 4                             // AvailableHmdCaps - DefaultTrackingCaps
    + SizeOf(ovrFovPort) * 2            // DefaultEyeFov
    + SizeOf(ovrFovPort) * 2            // MaxEyeFov
    + SizeOf(ovrSizei)                  // Resolution
    + 4                                 // DisplayRefreshRate
    }{$MESSAGE Error 'ovrHmdDesc size mismatch'}{$IFEND}
{$ENDIF}

implementation

const
  libovrdll = 'LibOVR.dll';

function ovr_Initialize; external libovrdll;
procedure ovr_Shutdown; external libovrdll;
procedure ovr_GetLastErrorInfo; external libovrdll;
function ovr_GetVersionString; external libovrdll;
function ovr_TraceMessage; external libovrdll;
function ovr_GetHmdDesc; external libovrdll;
function ovr_GetTrackerCount; external libovrdll;
function ovr_GetTrackerDesc; external libovrdll;
function ovr_Create; external libovrdll;
procedure ovr_Destroy; external libovrdll;
function ovr_GetSessionStatus; external libovrdll;
function ovr_SetTrackingOriginType; external libovrdll;
function ovr_GetTrackingOriginType; external libovrdll;
function ovr_RecenterTrackingOrigin; external libovrdll;
procedure ovr_ClearShouldRecenterFlag; external libovrdll;
function ovr_GetTrackingState; external libovrdll;
function ovr_GetTrackerPose; external libovrdll;
function ovr_GetInputState; external libovrdll;
function ovr_GetConnectedControllerTypes; external libovrdll;
function ovr_SetControllerVibration; external libovrdll;
function ovr_GetTextureSwapChainLength; external libovrdll;
function ovr_GetTextureSwapChainCurrentIndex; external libovrdll;
function ovr_GetTextureSwapChainDesc; external libovrdll;
function ovr_CommitTextureSwapChain; external libovrdll;
procedure ovr_DestroyTextureSwapChain; external libovrdll;
procedure ovr_DestroyMirrorTexture; external libovrdll;
function ovr_GetFovTextureSize; external libovrdll;
function ovr_GetRenderDesc; external libovrdll;
function ovr_SubmitFrame; external libovrdll;
function ovr_GetPredictedDisplayTime; external libovrdll;
function ovr_GetTimeInSeconds; external libovrdll;
function ovr_GetBool; external libovrdll;
function ovr_SetBool; external libovrdll;
function ovr_GetInt; external libovrdll;
function ovr_SetInt; external libovrdll;
function ovr_GetFloat; external libovrdll;
function ovr_SetFloat; external libovrdll;
function ovr_GetFloatArray; external libovrdll;
function ovr_SetFloatArray; external libovrdll;
function ovr_GetString; external libovrdll;
function ovr_SetString; external libovrdll;

end.